"""acting.py action module

"""
#print("module {0}".format(__name__))

import time
import struct
from collections import deque
from collections.abc import Mapping
from functools import wraps
import inspect
import copy
try:
    from itertools import izip
except ImportError: #python 3 zip is same as izip
    izip = zip

from ..aid.sixing import *
from .globaling import AUX, MOOT, INDENT_ADD, ScheduleNames, TRANSIT, ActionSubContextNames
from ..aid.odicting import odict

from ..aid.classing import metaclassify, nonStringIterable
from ..aid.aiding import nameToPath
from . import excepting
from . import registering
from . import storing
from . import framing

from ..aid.consoling import getConsole
console = getConsole()


#Class definitions

class Act(object):
    """ Container class for actor resolve time initialization and runtime operation """
    __slots__ = ('frame', 'context', 'act', 'actor',
                 'registrar', 'inits', 'ioinits', 'parms',
                 'prerefs', 'human', 'count', 'inode')

    def __init__(self, frame=None, context=None, act=None, actor=None,
                 registrar=None, inits=None, ioinits=None, parms=None,
                 prerefs=None, human='', count=0, inode=None, **kwa ):
        """
        Initialization method for instance.

        Attributes:
            .frame = ref to Frame holding this Act
            .context = name string of actioning context set when added to frame
            .act = ref to super Act if self is embedded in another Act such as Need
            .actor = ref to Actor instance or actor name to be resolved
                    call performs instances method action
            .registrar = ref to Actor class holding .Registry
            .parms = dictionary of keyword arguments for Actor instance call
            .inits = dictionary of arguments to Actor.__init__()
            .ioinits = dictionary of arguments to Actor._initio()
            .prerefs = dictionary of init, ioinit, and parm refs to resolve
            .human = human friendly version of action declaration
            .count = line count in floscript of action declaration
            .inode = actor level inode for ioinits if any

        Share path resolution
            act.resolve aggregates ioinits from registry,  act.ioinits, and act.prerefs
            act.resolve calls .actor._initio with ionits as parameters to get iois
            then for each ioi act.resolve calls .resolvePath to get share or node
            and then assigns share/node to to attribute/parameter of .actor

        """
        super(Act,self).__init__(**kwa) # in case of MRO

        self.frame = frame
        self.context = context
        self.act = act # should normally not be set until resolve time
        self.actor = actor #callable instance performs action
        self.registrar = registrar or Actor
        self.parms = parms if parms is not None else odict() # parms must always be not None
        self.inits = inits if inits else None # store None if empty dict
        self.ioinits = ioinits if ioinits else None # store None if empty dict
        self.prerefs = prerefs if prerefs else None # store None if empty dict
        self.human = human  # human readable version of FloScript declaration
        self.count = count  # line number or count of FloScript declaration
        self.inode = inode  # inode None means ignore

    def clone(self):
        """ Return clone of self in support of framer cloning
            clones is dict whose items keys are original framer names
            and values are duples of (original,clone) framer references
        """
        if isinstance(self.frame, framing.Frame):
            msg = ("CloneError: Attempting to clone resolved frame link"
                "  '{0}'.".format(self.frame.name))
            raise excepting.CloneError(msg)

        if self.act:
            msg = ("CloneError: Attempting to clone resolved act link"
                   "  '{0}'.".format(self.act))
            raise excepting.CloneError(msg)

        if isinstance(self.actor, Actor):
            msg = ("CloneError: Attempting to clone resolved actor"
                "  '{0}'.".format(self.actor.name))
            raise excepting.CloneError(msg)

        # frame and context for Act set when add Act to frame later
        clone = copy.deepcopy(self)
        return clone

    def __call__(self): #make Act instance callable as function
        """ Define act as callable object """
        return (self.actor(**self.parms))

    def expose(self):
        """ Show attributes"""
        console.terse("Act Actor {0} Parms {1} in Frame {2} Context {3} SuperAct {4}\n".format(
                    self.actor, self.parms, self.frame, self.context, self.act))
        if self.actor:
            self.actor._expose()

    def resolve(self, **kwa):
        """ Resolve .frame attribute and .actor. Cause .actor to resolve its parms """
        if self.act: # this is sub act of another act self.act so get super act context
            self.frame = self.act.frame
            self.context = self.act.context
        self.frame = framing.resolveFrame(self.frame,
                                          who="Act for {0}".format(self.actor),
                                          desc="act's",
                                          human=self.human,
                                          count=self.count)

        if not isinstance(self.actor, Actor): # Need to resolve .actor
            # .actor is currently unresolved name string for actor
            actor, inits, ioinits, parms = self.registrar.__fetch__(self.actor)

            if self.prerefs: # preinits inits dict items  'do qua'
                # each key is share src path, and value is list of src fields
                for src, fields in self.prerefs.get('inits', {}).items():
                    src = self.resolvePath(ipath=src, warn=True) # now a share
                    if not fields:  # default is use existing fields
                        fields = self._prepareSrcFields(src, fields)
                    for field in fields:  # requires src pre inited
                        if field in src:  # only update if src has field
                            inits[field] = src[field]
            inits.update(self.inits or odict())  # 'do cum' overrides qua, self.inits may be None
            if 'name' not in inits:
                inits['name'] = self.actor  # as yet unresolved name string
            inits['store'] = self.frame.store
            inits['act'] = self
            self.actor = actor = actor(**inits) # instantiate and convert

            if self.prerefs: # preinits parms dict items 'do from'
                # each key is share src path, and value is list of src fields
                for src, fields in self.prerefs.get('parms', {}).items():
                    src = self.resolvePath(ipath=src, warn=True) # now a share
                    if not fields:  # default is use existing fields
                        fields = self._prepareSrcFields(src, fields)
                    for field in fields:  # requires src pre inited
                        if field in src:  # only update if src has field
                            parms[field] = src[field]
            parms.update(self.parms or odict())  # 'do with' overrides from, self.parms may be None

            #  compute .inode
            self.inode = (self.ioinits or odict()).get('inode', ioinits.get('inode', self.inode))
            if self.inode is not None and not isinstance(self.inode, (str, bytes)):
                raise ValueError("Nonstring inode '{0}'".format(self.inode))
            if self.inode and not self.inode.endswith("."):  # ensure node not share path
                self.inode = "{0}.".format(self.inode)

            rioinits = odict(ioinits)  # save copy registry ioinits to keep ival iown
            if self.prerefs and 'ioinits' in self.prerefs:  # preinit ioinits dict items  'do for'
                # each key is share src path, and value is list of src fields
                if self.inode is None:  # since prerefs then inode may not be None
                    self.inode = ""
                for src, fields in self.prerefs['ioinits'].items():  # now resolve new src paths
                    src = self.resolvePath(ipath=src, warn=True)  # now a share
                    if not fields:  # default is use existing fields
                        fields = self._prepareSrcFields(src, fields)
                    for field in fields:  # assumes each src fld value pre inited with ipath
                        if field in src and field != "inode":  # only update if src has field
                            ioinits[field] = src[field] # clobbers ioinit odict() with ipath
            ioinits.update(self.ioinits or odict())  # 'do per', 'via inode', self.ioinits may be None
            for key, val in rioinits.items():  # restore registry ival iown defaults
                if key in ioinits:  # check ipath overridden by for or per
                    # if new val not mapping and old val is mapping
                    if not isinstance(ioinits[key], Mapping) and isinstance(val, Mapping):
                        val = val.copy()  # make copy so don't pollute registry
                        val['ipath'] = ioinits[key]  # replace old ipath witn new
                        ioinits[key] = val # restore default ival iown


            if ioinits:
                iois = actor._initio(ioinits)  # .inode may be changed in here
                if iois:
                    for key, ioi in iois.items():
                        if key == "inode":  # compute the final inode
                            share = actor._resolvePath(ipath="", # empty to force
                                                       ival=ioi.get('ival'),
                                                       iown=ioi.get('iown'))
                        else:
                            share = actor._resolvePath(ipath=ioi['ipath'],
                                                       ival=ioi.get('ival'),
                                                       iown=ioi.get('iown'))
                        if actor._Parametric:
                            if key in parms:
                                msg = "ResolveError: Parm and Ioi with same name"
                                raise excepting.ResolveError(msg,
                                                             key,
                                                             self,
                                                             self.human,
                                                             self.count)
                            parms[key] = share
                        else:
                            if hasattr(actor, key):
                                msg = "ResolveError: Attribute and Ioi with same name"
                                raise excepting.ResolveError(msg,
                                                             key,
                                                             self,
                                                             self.human,
                                                             self.count)
                            setattr(actor, key, share)
            self.parms = parms
            self.parms.update(self.actor._resolve(**self.parms)) # resolve sub acts
            self.actor._prepare(**self.parms)

    def resolvePath(self, ipath, ival=None, iown=None, warn=False):
        """
        Returns resolved Share or Node instance from ipath
        ipath may be path name of share or node
        or reference to Share or Node instance

        This method resolves inline pathname strings into share and node references
        at resolve time.

        ival is optional value for share
        iown is optional ownership if True then overwrite if exists
               otherwise do not overwrite

        if warn then will complain if the share is created.

        When ipath is empty then resolvePath returns the effective context node
        path given by .inode and nested frame inodes and framer inodes

        It allows for substitution into ipath of
        inode, framer,  main framer, frame, main frame, or actor relative names.
        So that lexically relative pathnames can
        be dynamically resolved in support of framer cloning.
        It assumes that any implied variants have been reconciled.
        If .actor is not yet instantiated then raises exception if ipath
           uses actor relative addressing. This may occur for init prerefs.

        When ipath is a string:  (not a Node or Share reference)
            the following syntax is used:

            If the path name starts with a leading '.' dot then path name is
            fully reconciled and no contextual substitutions are to be applied.

            If the path name begins with 'me' then the path is framer inode relative

            Otherwise make subsitutions in pathname strings that begin
            with 'framer.'
                Substitute for special path part 'framer' with names of 'me' or 'main'
                Substitute for special path part 'frame' with names  of 'me' or 'main'
                Substitute for special path part 'actor' with name of 'me'

                'me' indicates substitute the current framer, frame, or actor name
                respectively.

                'main' indicates substitute the current .main.frame.name or
                .main.frame.framer.name respectively obtained from aux .main link



        If ipath ends with a '.' then it resolves to a node
        Otherwise if not trailing dot it resolves to a share

        When  ipath is a pathname string that resolves to a Share and ival is not None
        Then ival is used to initialize the share values.
            ival should be a share initializer:
               valid initializers are:
                   a dict of fields and values
                   a list of duples, each a (key, value) item

            If own is True then .update(ival) is used
            Otherwise .create(ival) is used

        Requires that:
           self.store exist

           The following have already been resolved:
               self.frame
               self.frame.framer
               self.frame.framer.main
               self.frame.framer.main.framer

           If actor relative addressing is used then:
              self.actor is resolved
              self.actor.name is not empty

        Any actor could have ioinits saved for them in their ._act  Act
        If ionits exist then then an inode should be part of the ioinits
        Currently only the Do verb actors have ioinits. The inode is one of them.
        The ioinits come from two sources. The registry and the do verb per and for
        clauses

        _act.ioinits are the do and per clauses
        registry ioints are from the Actor.Ionints class variable

        So the .inode attribute indicates if the ipath being resolved is for
        an ioinit or not.  If .inode is None it is not for an ioinit.
        Otherwise even if .inode is "" then ipath comes from ioinit and
        path resolution should handle inode logic.

        This would puts all the framer inode prepend
        logic in one place and make it easier to climb the  frame and aux outline
        """
        if not (isinstance(ipath, storing.Share) or isinstance(ipath, storing.Node)): # must be pathname
            parts = ipath.split('.') if ipath else []  # [] if ipath is empty
            noded = True if not parts else False  # return node path prefix context since empty ipath
            if not parts or parts and parts[0]:  # empty or not empty and not absolute so do relative substitutions
                if not self.frame:
                    raise excepting.ResolveError("ResolveError: Missing frame context"
                                             " to resolve relative pathname.", ipath, self,
                                             self.human, self.count)
                if not self.frame.framer:
                    raise excepting.ResolveError("ResolveError: Missing framer context"
                                                 " to resolve relative pathname.", ipath, self.frame,
                                                 self.human, self.count)

                framer = self.frame.framer
                fparts = framer.inode.rstrip(".").split(".") if framer.inode else []
                main = framer.main  # main frame
                mainer = main.framer if main else None  # main frame's framer
                # if mainer and not (absolute or framer relative)
                # then fromer is aux framer so walk up mains processing inodes
                # need to do this first because used by both inode and ipath logic below
                while mainer and (not fparts or fparts[0] not in ("", "framer")):
                    # four aux inode cases:
                    # absolute: fparts[0] = "",
                    # framer relative: fparts[0] == "framer",
                    # empty: fparts == [],
                    # me relative: fparts[0] == "me",
                    # relative: otherwise

                    if fparts and fparts[0] == "me":
                        del fparts[0]  # me relative
                    else: # process over frame via inode
                        while main and (not fparts or fparts[0] not in ("", "framer")):
                            if main.inode:
                                fparts = main.inode.rstrip(".").split(".") + fparts  # prepend
                                if fparts and fparts[0] == "me":  # me relative
                                    del fparts[0]
                                    break  # skip any further over frames to main frame
                            main = main.over  # walk up to over frame if any
                        if fparts and fparts[0] in ("", "framer"):
                            break

                    if mainer.inode:
                        fparts = mainer.inode.rstrip(".").split(".") + fparts  # prepend
                    main = mainer.main  # main framer's main frame walk up one level
                    mainer = main.framer if main else None  # walk up one level

                if fparts and fparts[0] == "me" :
                    del fparts[0]  # me not allowed as initial inode

                # "" == ".".join([]) == ".".join([""])

                # over frame inode processing
                frame = self.frame # start at current frame
                oparts = frame.inode.rstrip(".").split(".") if frame.inode else []
                while frame and (not oparts or oparts[0] not in ("", "framer")):
                    if oparts and oparts[0] == "me":  # me relative
                        del oparts[0]
                        break  # skip any further over frames
                    frame = frame.over
                    if frame and frame.inode:
                        oparts = frame.inode.rstrip(".").split(".") + oparts

                # already know from above parts not absolute
                if (self.inode is not None and (not parts or
                        parts[0] not in ("framer", "me"))):
                    # prepend act .inode logic
                    iparts = self.inode.rstrip(".").split(".") if self.inode else []
                    if not iparts and not oparts and not fparts:  # use default inode
                        iparts = "framer.me.frame.me.actor.me".split(".")
                    parts = iparts + parts  # prepend inode parts

                if not parts or parts[0] not in ("", "framer"):  # not absolute or framer relative
                    if parts and parts[0] == 'me':  # skip frame.inode processing
                        del parts[0]  # parts could be empty now
                    else:  # prepend over frame inodes
                        parts = oparts + parts
                    if not parts or parts[0] not in ("", "framer"):
                        parts = fparts + parts  # prepend fparts framer inodes

            if parts and parts[0]:  # not absolute so do relative substitutions
                if parts[0] == 'framer':  #  framer relative addressing
                    if parts[1] == 'me': # current framer
                        parts[1] = self.frame.framer.name
                    elif parts[1] == 'main': # current main framer
                        if not self.frame.framer.main:
                            raise excepting.ResolveError("ResolveError: Missing main framer"
                                " context to resolve relative pathname.", ipath,
                                self.frame.framer,
                                self.human, self.count)
                        parts[1] = self.frame.framer.main.framer.name

                    if (len(parts) >= 3):
                        if parts[2] == 'frame':
                            if parts[3] == 'me':
                                parts[3] = self.frame.name
                            elif parts[3] == 'main':
                                if not self.frame.framer.main:
                                    raise excepting.ResolveError("ResolveError: "
                                        "Missing main frame context to resolve "
                                        "relative pathname.", ipath,
                                        self.frame.framer,
                                        self.human, self.count)
                                parts[3] = self.frame.framer.main.name
                            if (len(parts) >= 5 ) and parts[4] == 'actor':
                                if parts[5] == 'me': # slice insert multiple parts
                                    if (not isinstance(self.actor, Actor)
                                            or not self.actor.name): # unresolved actor
                                        raise excepting.ResolveError("ResolveError: Unresolved actor"
                                            " context to resolve relative pathname.", ipath, self,
                                                        self.human, self.count)
                                    parts[5:6] = nameToPath(self.actor.name).lstrip('.').rstrip('.').split('.')
                        elif parts[2] == 'actor':
                            if parts[3] == 'me': # slice insert multiple parts
                                if (not isinstance(self.actor, Actor)
                                        or not self.actor.name):
                                    raise excepting.ResolveError("ResolveError: Unresolved actor"
                                        " context to resolve relative pathname.", ipath, self,
                                               self.human, self.count)
                                parts[3:4] = nameToPath(self.actor.name).lstrip('.').rstrip('.').split('.')

            ipath = '.'.join(parts)

            if noded and ipath and not ipath.endswith("."):
                ipath = "{0}.".format(ipath)

            if not self.frame.store:
                raise excepting.ResolveError("ResolveError: Missing store context"
                                " to resolve relative pathname.", ipath, self,
                                self.human, self.count)

            if ipath.endswith('.'): # Node not Share
                ipath = self.frame.store.createNode(ipath.rstrip('.'))
                if warn:
                    console.profuse( "     Warning: Non-existent node '{0}' "
                                        "... creating anyway\n".format(ipath))
            else: # Share
                ipath = self.frame.store.create(ipath)
                if ival is not None:
                    if iown:
                        ipath.update(ival)
                    else:
                        ipath.create(ival)
                if warn:
                    console.profuse( "     Warning: Non-existent node '{0}' "
                                     "... creating anyway\n".format(ipath))

        return ipath

    def _prepareSrcFields(self, src, fields):
        """
        Prepares and verifys list field names fields in share src
           handles default conditions when fields is empty

           src is share
           fields is list of field names

        """
        if not fields: # empty source fields so assign defaults
            if src:
                if 'value' in src:
                    fields = ['value'] #use value field
                else: #use pre-existing source fields
                    fields = src.keys()
            else: # empty src
                fields = ['value'] #use value field
        return fields

class Nact(Act):
    """ Negating Act used for actor needs to give Not Need
    """
    __slots__ = ('frame', 'context', 'act',
                 'actor', 'registrar', 'parms', 'inits', 'ioinits',
                 'human', 'count')

    def __init__(self, **kwa ):
        """ Initialization method for instance. """
        super(Nact,self).__init__(**kwa)

    def __call__(self): #make Act instance callable as function
        """Define act as callable object
           Negate the output
        """
        return not (self.actor(**self.parms))

    def expose(self):
        """ Show attributes """
        console.terse("Nact Actor {0} Parms {1} in Frame {2} Context {3} SuperAct {4}\n".format(
                    self.actor, self.parms, self.frame, self.context, self.act))
        if self.actor:
            self.actor._expose()

class SideAct(Act):
    """ Anciliary act to a main Act/Actor used to call a different 'action' method
        of the Main act.actor in support of combined activity such as restarting
        an action.

        SideActs are created in the .resolve method of an Actor and assume that
        all attributes and parms have already been resolved.

        Unique Attributes:
            .action = name of method to call in .actor
    """
    __slots__ = ('action')

    def __init__(self, action='', **kwa ):
        """ Initialization method for instance. """
        super(SideAct, self).__init__(**kwa)
        self.action = action

    def __call__(self): #make Act instance callable as function
        """ Define call method named .action of .actor """
        return (getattr(self.actor, self.action)(**self.parms))

    def resolve(self, **kwa):
        """ Assumes all has been resolved.
            Check for valid action
        """
        if not isinstance(self.actor, Actor): # Woops
            msg = "ResolveError: Unresolved actor"
            raise excepting.ResolveError(msg, self.actor, self)

        if not self.action : # Woops
            msg = "ResolveError: Empty action"
            raise excepting.ResolveError(msg,
                                         self.action,
                                         self,
                                         self.human,
                                         self.count)

        if not getattr(self.actor, self.action, None):
            msg = "ResolveError: Missing action in actor"
            raise excepting.ResolveError(msg,
                                         self.action,
                                         self,
                                         self.human,
                                         self.count)

def actify(name, base=None, registry=None, inits=None, ioinits=None, parms=None,
             parametric=None):
    """ Parametrized decorator function that converts the decorated function
        into an Actor sub class with .action method and with class name that
        is name and registers the new subclass in the registry under name as
        a subclass of base. The default base is Actor.

        The parameters  registry, parametric, inits, ioinits, and parms if provided,
        are used to create the class attributes for the new subclass

    """
    base = base or Actor
    if not issubclass(base, Actor):
        msg = "Base class '{0}' not subclass of Actor".format(base)
        raise excepting.RegisterError(msg)

    attrs = odict()
    if registry:
        attrs['Registry'] = odict()
    if parametric is not None:
        attrs['_Parametric'] = True if parametric else False
    if inits:
        attrs['Inits'] = odict(inits)
    if ioinits:
        attrs['Ioinits'] = odict(ioinits)
    if parms:
        attrs['Parms'] = odict(parms)
    cls = type(name, (base, ), attrs )

    def implicit(func):
        @wraps(func)
        def inner(*pa, **kwa):
            return func(*pa, **kwa)
        cls.action = inner
        return inner
    return implicit

actorify = actify  # alias for backwards compatibility

@metaclassify(registering.RegisterType) # python2or3 compatible
class Actor(object):
    """ Actor Base Class
        Has Actor specific Registry of classes
    """
    Registry = odict() # Actor Registry
    Inits = odict() # class defaults support for
    Ioinits = odict() # class defaults
    Parms = odict() # class defaults
    _Parametric = True # Convert iois to action method parameters if Truthy
    __slots__ = ('name', 'store', '_act')

    def __init__(self, name='', store=None, act=None, **kwa ):
        """
        Initialization method for instance.

        Instance attributes
            .name = name string for Actor variant in class Registry
            .store = reference to shared data Store
            ._act = reference to containing Act

        If subclass has init need to call super
        super(SubClassName, self).__init__(**kwa)

        """
        self.name = name
        if store is not None:
            if  not isinstance(store, storing.Store):
                raise ValueError("Not store {0}".format(store))
            self.store = store
        self._act = act


    def __call__(self, **kwa):
        """ run .action  """
        return self.action(**kwa)

    def action(self, **kwa):
        """Action called by Actor. Should override in subclass."""
        console.profuse("Actioning {0} in {1} of {2} with {3}\n".format(self.name,
                                                                        self._act.frame.name,
                                                                        self._act.frame.framer.name,
                                                                        str(kwa)))
        pass

    def _expose(self):
        """Show Actor."""
        console.terse("Actor {0}".format(self.name))

    def _resolve(self, **kwa):
        """ Return updated parms
            Extend in subclass to resolve specific kwa items that are links or
            share refs and update parms
        """
        parms = odict()

        return parms

    def _initio(self, ioinits):
        """
        Compute initializations for ioflo shares from ioinits (odict or item list)

        The 'inode' item in ioinits is special. Act.resolve extracts 'inode' from
        ioinits and assigns to ._act.inode before ._initio is called.

        This implements a generic Actor interface protocol for associating the
        io data flow shares to the Actor.

        The ._act.inode attribute holds the pathname string of the default
        data store node where shares associated with the Actor instance may
        be placed when relative addressing is used.
        It is also the default data store context
        for share references by Doer iois. The inode on an actor my be set
        explicitly in the Ioinit class variable or in the actify (doify) decorator
        or by the via clause of a Doer

        The values of the items in the **kwa argument may be either strings or
        mappings

        For each key,val in ioinits.items() there are the following 2 forms for val:

        1- string:
           ipath = pathnamestring
        2- dict of items (mapping) of form:
            {
                ipath: "pathnamestring", (optional)
                ival: initial value, (optional)
                iown: truthy, (optional)
            }

        In either case, three init values are produced, these are:
            ipath, ival, iown,
        Missing init values from ipath, ival, iown will be assigned a
        default value as per the rules below:

        For each ioinits item (key, val) besides the inode
            key is the name of the associated instance attribute or action parm.
            Extract ipath, ival, iown from val

            Shares are initialized with mappings passed into share.create or
            share.update. So to assign ival to share.value pass into share.create
            or share.update a mapping of the form {'value': ival} whereas
            passing in an empty mapping does nothing.

            If ipath not provided
                ipath is the default path inode.key
            Else ipath is provided
                If ipath starts with dot "." Then absolute path
                Else ipath does not start with dot "." Then relative path from inode

                If ipath ends with dot Then the path is to a node not share
                        node ref is created and remaining init values are ignored


            If ival not provided
                 set ival to empty mapping which when passed to share.create
                 will not change share.value
            Else ival provided
                If ival is an empty Mapping Then
                    assign a shallow copy of ival to share.value by passing
                    in {value: ival (copy)} to share.create/.update
                Else If ival is a non-string iterable and not a mapping
                    assign a shallow copy of ival to share.value by passing
                    in {value: ival (copy)} to share.create/.update
                Else If ival is a non-empty Mapping Then
                    Each item in ival is assigned as a field, value pair in the share
                    by passing ival directly into share.create or share.update
                    This means there is no way to init a share.value to a non empty mapping
                    It is possible to init a share.value to an empty mapping see below
                Else
                    assign ival to share.value by by passing
                    in {value: ival} to share.create or share.update

            Create share with pathname given by ipath

            If iown Then
                init share with ival value using update
            Else
                init share with ival value using create ((change if not exist))

        """
        ioinits = odict(ioinits)  # make copy also allows for item list as parameter
        #inode = ioinits.get('inode', '')

        #if inode and not inode.endswith('.'):  # ensure is a node not share path
            #inode = "{0}.".format(inode)

        if self._act.inode is None:  # inode may not be None if ioinits
            self._act.inode = ""

        iois = odict()
        ioi = odict(ipath=self._act.inode)  # create ioi for the inode
        iois['inode'] = ioi  # inode path resolves so actor has reference to its own inode
        for key, val in ioinits.items():  # assumes keys are basestrings
            if key == 'inode':  # skip inode
                continue

            if val is None:  # use default assume no way to be inadvertent
                val = ""  # empty string for val defaults ipath to same as key

            if isinstance(val, (str, bytes)):
                ipath = val
                iown = None
                ival = odict() # effectively will not change share
            elif isinstance(val, Mapping): # dictionary
                ipath = val.get('ipath', '')
                iown = val.get('iown')

                if not 'ival' in val:
                    ival = odict() # effectively will not change share
                else:
                    ival = val['ival']
                    if isinstance(ival, Mapping):
                        if not ival: #empty mapping
                            ival = odict(value=copy.copy(ival)) #make copy so each instance unique
                        # otherwise don't change since ival is non-empty mapping
                    elif nonStringIterable(ival): # not mapping and nonStringIterable
                        ival = odict(value=copy.copy(ival))
                    else:
                        ival = odict(value=ival)
            else:
                raise ValueError("Bad ioinit for key '{0}' with value '{1}'".format(key, val))

            # inode is prepended in act.resolvePath

            if not ipath:
                ipath = key  # when ipath empty create default from key

            ioi = odict(ipath=ipath, ival=ival, iown=iown)
            iois[key] = ioi

        return iois  # inode prepended in act.resolvePath

    def _prepare(self, **kwa):
        """ Base method to be overriden in sub classes. Perform post initio setup
            after all parms and ioi parms or attributes have been created
            kwa usually parms.

            If one is using this method consider refactoring into two different
            behaviors
        """
        pass

    def _resolvePath(self, ipath, ival=None, iown=None, warn=False):
        """ Returns resolved Share or Node instance from ipath
            Calls self._act.resolvePath()
            See doc string from Act.resolvePath for detailed description of
            functionality

            Requires that
               self._act exist
        """
        if not (isinstance(ipath, storing.Share) or isinstance(ipath, storing.Node)): # must be pathname
            if not self._act:
                raise excepting.ResolveError("ResolveError: Missing act context"
                        " to resolve relative pathname.", ipath, self,
                        self._act.human, self._act.count)
            ipath = self._act.resolvePath(ipath=ipath,
                                          ival=ival,
                                          iown=iown,
                                          warn=warn)

        return ipath

    def _prepareDstFields(self, srcFields, dst, dstFields):
        """
        Prepares  for a transfer of data
           from srcFields
           to dstFields in dst
        Handles default conditions when fields are empty
            srcFields is list of field names
            dst is share
            dstFields is list of field names

        Ensure Builder.prepareDataDstFields is similar
        """
        if not dstFields: #no destinationField so use default rules
            if 'value' in dst:
                dstFields = ['value'] #use value field

            else: #use srcFields
                dstFields = srcFields

        self._verifyShareFields(dst, dstFields)

        if len(srcFields) != len(dstFields):
            msg = ("ResolveError: Unequal number of source = {0} and "
                   "destination = {2} fields".format(srcFields, dstFields))
            raise excepting.ResolveError(msg,
                                         self.name,
                                         '',
                                         self._act.human,
                                         self._act.count)

        for dstField, srcField in izip(dstFields, srcFields):
            if (dstField != srcField) and (srcField != 'value'):
                console.profuse("     Warning: Field names mismatch. '{0}' in {1} "
                                "from '{2}' ... creating anyway".format(
                                  dstField, dst.name, srcField))

        #create any non existent destination fields
        for field in dstFields: #use destination fields for destination data
            if field not in dst:
                console.profuse("     Warning: Transfer into non-existent field '{0}' in "
                       "share {1} ... creating anyway\n".format(field, dst.name))
                dst[field] = None #create

        return dstFields

    def _prepareSrcDstFields(self, src, srcFields, dst, dstFields):
        """
        Prepares and verifys a transfer of data
           from srcFields in src
           to dstFields in dst
        Handles default conditions when fields are empty
           src and dst are shares
           srcFields and dstFields are lists

        Ensure Builder.prepareSrcDstFields is the same
        """
        if not srcFields: # empty source fields so assign defaults
            if src:
                if 'value' in src:
                    srcFields = ['value'] #use value field
                elif dstFields: #use destination fields for source fields
                    srcFields = dstFields
                else: #use pre-existing source fields
                    srcFields = src.keys()
            else: # empty src
                srcFields = ['value'] #use value field

        self._verifyShareFields(src, srcFields)

        if not dstFields: #no destination fields so assign defaults
            if 'value' in dst:
                dstFields = ['value'] #use value field
            else: #use source fields for destination fields
                dstFields = srcFields

        self._verifyShareFields(dst, dstFields)

        if len(srcFields) != len(dstFields):
            msg = ("ResolveError: Unequal number of fields, source = {0} and"
                  " destination={1)".format(srcFields, dstFields))
            raise excepting.ResolveError(msg,
                                         self.name,
                                         '',
                                         self._act.human,
                                         self._act.count)

        for dstField, srcField in izip(dstFields, srcFields):
            if (dstField != srcField) and (srcField != 'value'):
                console.profuse("     Warning: Field names mismatch. '{0}' in {1} "
                                "from '{2}' in {3}  ... creating anyway".format(
                                    dstField, dst.name, srcField, src.name))

        #create any non existent source or destination fields
        for field in srcFields: #use source fields for source data
            if field not in src:
                console.profuse("     Warning: Transfer from non-existent field '{0}' "
                        "in share {1} ... creating anyway".format(field, src.name))
                src[field] = None #create

        for field in dstFields: #use destination fields for destination data
            if field not in dst:
                console.profuse("     Warning: Transfer into non-existent field '{0}' "
                        "in share {1} ... creating anyway\n".format(field, dst.name))
                dst[field] = None #create

        return (srcFields, dstFields)

    def _verifyShareFields(self, share, fields):
        """
        Verify that updating fields in share won't violate the
           condition that when a share has field == 'value'
           it will be the only field

           fields is list of field names
           share is  share

        Raises exception if condition would be violated

        Ensure Builder.verifyShareFields is same
        """
        if (len(fields) > 1) and ('value' in fields):
            msg = "ResolveError: Field = 'value' within fields = '{0}'".format(fields)
            raise excepting.ResolveError(msg,
                                         self.name,
                                         '',
                                         self._act.human,
                                         self._act.count)

        if share: #does share have fields already
            for field in fields:
                if field not in share: #so this field could be added to share
                    if ('value' in share) or (field == 'value'):
                        msg = ("ResolveError: Candidate field '{0}' when "
                               "fields = '{1}' exist".format(field, share.keys()))
                        raise excepting.ResolveError(msg,
                                                     self.name,
                                                     '',
                                                     self._act.human,
                                                     self._act.count)

class Interrupter(Actor):
    """
    Interrupter Actor Class
    Interrupter is a base clase for all actor classes that interrupt normal precur
    processing and result in a change in the frame or the frame processing.

    This class must be subclassed. This is a convenience so can either use
      isinstance to test

    Specifically an Interrupter's action() method returns truthy when its action
    interrupts the normal frame processing.

    Examples are:
    Transiters which interrupt by changing to a new frame
    Suspenders which interrupt when the conditional aux condition is true and
       further processing of the frame and sub frames is stopped
    """
    def __init__(self,**kw ):
        """
        Initialization method for instance.

        Inherited Attributes:
            .name = name string for Actor variant in class Registry
            .store = reference to shared data Store
            ._act = reference to containing Act

        Attributes:
            ._tracts = list of references to transition acts for this Actor
                transit sub-context of precur context during segue

        """
        super(Interrupter,self).__init__(**kw)
        self._tracts = []

class Transiter(Interrupter):
    """Transiter Interrupter Class
       Transiter  is a special actor that performs transitions between frames

       Parameters
            needs = list of Act objects that are exit needs for this trans
            near = source frame of trans
            far = target frame of trans
            human = text version of transition

    """
    def _resolve(self, needs, near, far, human, **kwa):
        """Resolve any links in far and in associated parms for actors"""

        parms = super(Transiter, self)._resolve( **kwa)

        if near == 'me':
            near = self._act.frame

        parms['near'] = near = framing.resolveFrame(near,
                                                    who=self.name,
                                                    desc='near',
                                                    human=self._act.human,
                                                    count=self._act.count)


        if far == 'next':
            if not isinstance(near.next_, framing.Frame):
                raise excepting.ResolveError("ResolveError: Bad next frame",
                                             near.name, near.next_,
                                             self._act.human, self._act.count)
            far = near.next_

        elif far == 'me':
            far = near

        far = framing.resolveFrame(far,
                                   who=self.name,
                                   desc='far',
                                   human=self._act.human,
                                   count=self._act.count)

        parms['far'] = far

        for act in needs:
            act.act = self._act  # so act.resolve can ref self._act.frame.framer
            act.resolve()
            self._tracts.extend(act.actor._tracts)
            del act.actor._tracts[:]

        return parms


    def action(self, needs, near, far, human, **kw):
        """Action called by Actor  """

        framer = near.framer #to speed up

        console.profuse("Attempt segue from {0} to {1}\n".format(near.name, far.name))

        for act in needs:
            if not act(): #return None if not all true
                return None

        console.profuse("     active outline: {0}\n".format([frame.name for frame in framer.actives]))
        console.profuse("     far outline: {0}\n".format([frame.name for frame in far.outline]))

        #find uncommon entry and exit lists associated with transition
        #exits, enters = framing.Framer.Uncommon(framer.actives,far.outline)
        #find uncommon and common entry and exit lists associated with transition
        exits, enters, reexens = framing.Framer.ExEn(framer.actives, far)

        #check enters, if successful, perform transition
        if not framer.checkEnter(enters, exits):
            return None

        msg = "To: {0}<{1} at {2} Via: {3} ({4}) From: {5} after {6:0.3f}\n".format(
            framer.name, far.human, round(framer.store.stamp, 6), near.name, human,
            framer.human, framer.elapsed)
        console.terse(msg)

        console.profuse("     exits: {0}\n".format([frame.name for frame in exits]))
        console.profuse("     enters: {0}\n".format([frame.name for frame in enters]))
        console.profuse("     reexens: {0}\n".format([frame.name for frame in reexens]))

        for act in self._tracts:  # transit sub-context of segue precur
            act()

        framer.exit(exits) #exit uncommon frames in near outline reversed in place
        framer.rexit(reexens[:]) #make copy since reversed in place
        framer.renter(reexens)
        framer.enter(enters)
        framer.activate(active = far)
        return far

    def _expose(self):
        """      """
        console.terse("Transiter {0}\n".format(self.name))


class Suspender(Interrupter):
    """Suspender Interrupter Class
       Suspender  is a special actor that performs a conditional auxiliary
          which truncates the active outline effectively suspended the truncated
          frames

       Parameters
            needs = list of Act objects that are exit needs for this trans
            main = source frame of trans
            aux = target frame of trans
            human = text version of transition

    """
    def __init__(self,**kw ):
        """Initialization method for instance. """
        super(Suspender,self).__init__(**kw)

    def _resolve(self, needs, main, aux, human, **kwa):
        """Resolve any links aux and in associated parms for actors"""
        parms = super(Suspender, self)._resolve( **kwa)

        if main == 'me':
            main = self._act.frame

        parms['main'] = main = framing.resolveFrame(main,
                                                    who=main,
                                                    desc='main',
                                                    human=self._act.human,
                                                    count=self._act.count)

        parms['aux'] = aux = framing.resolveFramer(aux,
                                                   who=main.name,
                                                   desc='aux',
                                                   contexts=[AUX],
                                                   human=self._act.human,
                                                   count=self._act.count)

        for act in needs:
            act.act = self._act
            act.resolve()
            self._tracts.extend(act.actor._tracts)
            del act.actor._tracts[:]

        deActParms = odict(aux=aux)
        deAct = SideAct( actor=self,
                        parms=deActParms,
                        action='deactivize',
                        human=self._act.human,
                        count=self._act.count)
        self._act.frame.addExact(deAct)
        console.profuse("{0}Added exact {1} SideAct for {2} with {3} in {4}\n".format(
                INDENT_ADD, 'deactivize', self.name, deAct.parms, self._act.frame.name))
        deAct.resolve()

        return parms

    def action(self, needs, main, aux, human, **kw):
        """Action called by Actor  """

        framer = main.framer #to speed up

        if aux.done: #not active

            console.profuse("Attempt segue from {0} to aux {1}\n".format(main.name, aux.name))

            for act in needs:
                if not act(): #return None if not all true
                    return None

            #if aux.main: #see if aux still belongs to another frame
            #    return None

            # if aux.main is not None then it has not been released and so
            # we can't enter unless it is our act's frame
            if aux.main and (aux.main is not self._act.frame):
                console.concise("    Invalid aux '{0}' in use by another frame"
                        " '{1}'\n".format(aux.name, aux.main.name))
                return None

            if not aux.checkStart(): #performs entry checks
                return None

            msg = "To: {0} in {1}<{2} at {3} Via: {4} ({5}) From: {6} after {7:0.3f}\n".format(
                aux.name, framer.name, main.headHuman, round(framer.store.stamp, 6), main.name,
                human, framer.human, framer.elapsed)
            console.terse(msg)

            for act in self._tracts:  # transit sub-context of cond aux segue precur
                act()

            if aux.original:
                aux.main = main  #assign aux's main to this frame
            aux.enterAll() #starts at aux.first frame
            aux.recur()

            if aux.done: #if done after first iteration then clean up
                self.deactivate(aux)
                return None

            #truncate active outline to suspend lower frames
            framer.change(main.head, main.headHuman)
            return aux

        if not aux.done: #not done so active
            aux.segue()
            aux.recur()

            if aux.done: #if done after this iteraion clean up
                self.deactivate(aux)
                framer.reactivate()
                return None

            return aux


    def _expose(self):
        """      """
        console.terse("Suspender {0}\n".format(self.name))

    def deactivize(self, aux, **kwa):
        """ If not aux.done Then force deactivate. Used in exit action."""
        if not aux.done:
            console.profuse("{0} deactivate {1}\n".format(self.name, aux.name))
            self.deactivate(aux)

    def deactivate(self, aux):
        """Called by deactivator actor to cleanly exit      """
        console.profuse("Deactivating {0}\n".format(aux.name))

        aux.exitAll() # also sets .done = True
        if aux.original:
            aux.main = None



class Printer(Actor):
    """Printor Actor Class

       Printer is a special actor that just prints to console its message

    """
    def action(self, message, **kw):
        """Action called by Actor
        """
        #console.terse("{0} printer {1}\n".format(self.name, message))
        console.terse("*** {0} ***\n".format(message))

    def _expose(self):
        """   """
        console.terse("Printer {0}\n".format(self.name))

class Marker(Actor):
    """ Base class that sets up mark in provided share reference"""
    pass

class MarkerUpdate(Marker):
    """ MarkerUpdate Class

        MarkerUpdate is a special actor that acts on a share to mark the update by
            saving a copy of the last stamp

        MarkerUpdate works with NeedUpdate which does the comparison against the marked stamp.

        Builder at parse time when it encounters an NeedUpdate,
        creates the mark in the share and creates the appropriate MarkerUpdate
    """
    def action(self, share, marker, **kwa):
        """ Update mark in share
            Where share is reference to share and marker is unique name key of mark in
                share.marks odict
            Updates mark.stamp

            only one mark per marker per share is needed
        """
        console.profuse("{0} mark {1} in {2} on {3}\n".format(
            self.name, share.name, marker, 'update' ))

        mark = share.marks.get(marker)
        if mark:
            mark.stamp = self.store.stamp  # update stamp when marker runs
            if self._act.context == ActionSubContextNames[TRANSIT]:  # transit satisfied
                mark.used = mark.stamp

    def _expose(self):
        """   """
        console.terse("MarkerUpdate {0}\n".format(self.name))

class MarkerChange(Marker):
    """ MarkerChange Class

        MarkerChange is a special actor that acts on a share to mark save a copy
        of the data in the mark for the share.

        MarkerChange works with NeedChange which does the comparison with the mark

        Builder at parse time when it encounters a NeedChange,
        creates the mark in the share and creates the appropriate marker
    """
    def action(self, share, marker, **kwa):
        """ Update mark in share
            Where share is reference to share and marker is unique name key of mark in
                share.marks odict
            Updates mark.data

            only one mark per marker per share is needed
        """
        console.profuse("{0} mark {1} in {2} on {3}\n".format(
            self.name, share.name, marker, 'change' ))

        mark = share.marks.get(marker)
        if mark:
            mark.data = storing.Data(share.items())  # set date when marker runs

    def _expose(self):
        """   """
        console.terse("MarkerChange {0}\n".format(self.name))

class Rearer(Actor):
    """
    Rearer Actor Class
    Rearer is a special actor that clones a moot framer at runtime

       Parameters
            original = moot framer to be cloned
            clone = name of clone
            schedule = schedule kind of clone
            frame = frame to put auxiliary clone in

    """
    def _resolve(self, original, clone, schedule, frame, **kwa):
        """
        Resolve any links
        """
        parms = super(Rearer, self)._resolve( **kwa)

        parms['original'] = original = framing.resolveFramer(original,
                                                    who=self.name,
                                                    desc='original',
                                                    contexts=[MOOT],
                                                    human=self._act.human,
                                                    count=self._act.count)

        if schedule not in [AUX]:
            msg = ("ResolveError: Invalid schedule '{0}' for clone"
                  "of '{1}'".format(ScheduleNames.get(schedule, schedule),
                                  original.name))
            raise excepting.ResolveError(msg=msg,
                                         name=self.name,
                                         value=schedule,
                                         human=self._act.human,
                                         count=self._act.count)

        if schedule == AUX:  # only current framer
            parms['framer'] = framer = framing.resolveFramer(self._act.frame.framer,
                                            who=self._act.frame.name,
                                            desc='rear aux clone',
                                            contexts=[],
                                            human=self._act.human,
                                            count=self._act.count)

            if frame == 'me':  # cannot rear in current frame
                msg = ("ResolveError: Invalid frame 'me' for reared clone.")
                raise excepting.ResolveError(msg,
                                             name=clone,
                                             value=self.name,
                                             human=self._act.human,
                                             count=self._act.count)

            # frame required
            parms['frame'] = frame = framing.resolveFrameOfFramer(frame,
                                                                  framer,
                                                                  who=self._act.frame.name,
                                                                  desc='rear aux clone',
                                                                  human=self._act.human,
                                                                  count=self._act.count)

            if clone != 'mine':
                msg = "ResolveError: Aux insular clone name must be 'mine' not '{0}'".format(clone)
                raise excepting.ResolveError(msg,
                                             name=clone,
                                             value=self.name,
                                             human=self._act.human,
                                             count=self._act.count)

        else:
            msg = ("ResolveError: Invalid insular clone schedule '{0}' "
                                               "for {1}.".format(schedule, original))
            raise excepting.ResolveError(msg,
                                         name=clone,
                                         value=self.name,
                                         human=self._act.human,
                                         count=self._act.count)

        return parms

    def action(self, original, clone, schedule, frame, framer, **kw):
        """
        Action called by Actor

        Parameters:
           original is resolved original moot framer to be cloned
           clone is clone tag (not used currently)
           schedule is clone schedule
           frame is resolved frame for reared clone
           framer is resolved framer for frame for reared clone

        """

        console.profuse("         Cloning '{0}' as '{1}' be '{2}'\n".format(
                original.name, clone, ScheduleNames.get(schedule, schedule)))

        if schedule == AUX:
            if frame in self._act.frame.outline:
                console.terse("         Error: Cannot rear clone in own"
                              " '{0}' outline. {1} in line {2}\n".format(frame.name,
                                                                         self._act.human,
                                                                         self._act.count))
                return

            tag = framer.newAuxTag(base=original.tag)
            name = "_".join((framer.surname, tag))  # replace name with full name
            console.terse("         Rearing original '{0}' as aux insular clone"
                          " '{1}' in Frame '{2}' in Framer '{3}'"
                          "\n".format(original.name,
                                        name,
                                        frame.name,
                                        framer.name))
            clone = original.clone(name=name, tag=tag, schedule=schedule)
            clone.original = False  # main frame will be fixed
            clone.insular = True  #  local to this framer
            clone.razeable = True  # can be razed
            framer.auxes[tag] = clone
            frame.addAux(clone)
            clone.main = frame
            #self._act.frame.framer.assignFrameRegistry()  # setup for resolution
            self.store.house.presolvables.append(clone)
            self.store.house.presolvePresolvables()
            self.store.house.resolveResolvables()


class Razer(Actor):
    """
    Razer Actor Class
    Razer is a special actor that destroys a  framer clone at runtime

       Parameters
            who = auxiliary clones to be destroyed
            frame = frame holding clones to be destroyed

    """
    def _resolve(self, who, frame, **kwa):
        """
        Resolve any links
        """
        parms = super(Razer, self)._resolve( **kwa)


        parms['framer'] = framer = framing.resolveFramer(self._act.frame.framer,
                                        who=self._act.frame.name,
                                        desc='rear aux clone',
                                        contexts=[],
                                        human=self._act.human,
                                        count=self._act.count)

        if frame == 'me':  # cannot rear in current frame
            frame = self._act.frame

        # frame required
        parms['frame'] = frame = framing.resolveFrameOfFramer(frame,
                                                              framer,
                                                              who=self._act.frame.name,
                                                              desc='rear aux clone',
                                                              human=self._act.human,
                                                              count=self._act.count)

        return parms

    def action(self, who, frame, framer, **kw):
        """
        Action called by Actor

        Parameters:
           who is string describing which aux(es) to raze
           frame is frame with auxes to raze
           framer is framer of frame with auxes to raze
        """

        razeables = []
        if who == 'all':
            for aux in frame.auxes:
                if aux.insular and aux.razeable:
                    razeables.append(aux)

        elif who == 'first':
            for aux in frame.auxes:  # works even if .auxes empty
                if aux.insular and aux.razeable:
                    razeables.append(aux)
                    break

        elif who == 'last':  # works even if .auxes empty
            for aux in reversed(frame.auxes):
                if aux.insular and aux.razeable:
                    razeables.append(aux)
                    break

        for aux in razeables:
            console.concise("         Razing '{0}' in '{1}'\n".format(who, frame.name))
            aux.prune()
            frame.auxes.remove(aux)
            if aux.tag in framer.auxes:
                del framer.auxes[aux.tag]



# -*- coding: utf-8 -*-
"""
Unittests for serialing module
"""

import sys
import unittest

import os
import time
import tempfile
import shutil
import socket
import errno

from ioflo.aid.sixing import *
from ioflo.aid.consoling import getConsole
from ioflo.aio.serial import serialing

console = getConsole()

def setUpModule():
    console.reinit(verbosity=console.Wordage.concise)

def tearDownModule():
    console.reinit(verbosity=console.Wordage.concise)

class BasicTestCase(unittest.TestCase):
    """
    Test Case
    """

    def setUp(self):
        """

        """

    def tearDown(self):
        """

        """

    def testConsoleNb(self):
        """
        Test Class ConsoleNB
        """
        console.terse("{0}\n".format(self.testConsoleNb.__doc__))

        myconsole = serialing.ConsoleNb()
        result = myconsole.open()
        #self.assertIs(result, False)
        #self.assertIs(result, True)

        #cout = "Enter 'hello' and hit return: "
        #myconsole.put(cout)
        #cin = ''
        #while not cin:
            #cin = myconsole.getLine()
        #myconsole.put("You typed: " + cin)
        #self.assertEqual('hello\n', cin)

        myconsole.close()



def runOne(test):
    '''
    Unittest Runner
    '''
    test = BasicTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runSome():
    """ Unittest runner """
    tests =  []
    names = ['testConsoleNb',
            ]
    tests.extend(map(BasicTestCase, names))
    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)

def runAll():
    """ Unittest runner """
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(BasicTestCase))
    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:
    #runAll() #run all unittests
    runSome()#only run some

    #runOne('testClientAutoReconnect')





# -*- coding: utf-8 -*-
"""
Unittests
"""
from __future__ import absolute_import, division, print_function

import sys

import unittest

import os

from binascii import hexlify

# Import ioflo libs
from ioflo.aid.sixing import *
from ioflo.aid.byting import hexify, bytify, unbytify, packify, unpackify
from ioflo.aid import getConsole

console = getConsole()

from ioflo.aio.proto import devicing, stacking

def setUpModule():
    console.reinit(verbosity=console.Wordage.concise)

def tearDownModule():
    console.reinit(verbosity=console.Wordage.concise)

class BasicTestCase(unittest.TestCase):
    """
    Test Case
    """

    def setUp(self):
        """

        """
        console.reinit(verbosity=console.Wordage.profuse)

    def tearDown(self):
        """

        """
        console.reinit(verbosity=console.Wordage.concise)

    def testDevice(self):
        """
        Test Device class
        """
        console.terse("{0}\n".format(self.testDevice.__doc__))

        stack = stacking.Stack()
        self.assertIsInstance(stack.local, devicing.LocalDevice)
        self.assertEqual(stack.local.uid, 1)
        self.assertEqual(stack.local.name, "Device{0}".format(stack.local.uid))
        self.assertEqual(stack.local.ha, "")
        self.assertEqual(stack.local.kind, None)

        device = devicing.Device(stack=stack)
        self.assertEqual(device.stack, stack)
        self.assertEqual(device.uid, 2)  # stacks .local has uid 1
        self.assertEqual(device.name, "Device{0}".format(device.uid))
        self.assertEqual(device.ha, '')
        self.assertEqual(device.kind, None)

        stack = stacking.RemoteStack()
        self.assertIsInstance(stack.local, devicing.LocalDevice)
        self.assertEqual(stack.local.uid, 1)
        self.assertEqual(stack.local.name, "Device{0}".format(stack.local.uid))
        self.assertEqual(stack.local.ha, "")
        self.assertEqual(stack.local.kind, None)

        device = devicing.RemoteDevice(stack=stack)
        self.assertIs(device.stack, stack)
        self.assertEqual(device.uid, 2)
        self.assertEqual(device.name, "Device{0}".format(device.uid))
        self.assertEqual(device.ha, '')
        self.assertEqual(device.kind, None)


    def testUdpDevice(self):
        """
        Test UdpDevice class
        """
        console.terse("{0}\n".format(self.testUdpDevice.__doc__))

        stack = stacking.UdpStack()
        self.assertIsInstance(stack.local, devicing.IpLocalDevice)
        self.assertEqual(stack.local.uid, 1)
        self.assertEqual(stack.local.name, "Device{0}".format(stack.local.uid))
        self.assertEqual(stack.local.ha, ('127.0.0.1', stacking.UdpStack.Port))
        self.assertEqual(stack.local.kind, None)
        self.assertEqual(stack.aha, ('0.0.0.0', stacking.UdpStack.Port))

        device = devicing.IpDevice(stack=stack)
        self.assertEqual(device.stack, stack)
        self.assertEqual(device.uid, 2)  # stacks .local has uid 1
        self.assertEqual(device.name, "Device{0}".format(device.uid))
        self.assertEqual(device.ha, ('127.0.0.1',stacking.UdpStack.Port))
        self.assertEqual(device.kind, None)

        device = devicing.IpRemoteDevice(stack=stack)
        self.assertIs(device.stack, stack)
        self.assertEqual(device.uid, 3)
        self.assertEqual(device.name, "Device{0}".format(device.uid))
        self.assertEqual(device.ha, ('127.0.0.1', stacking.UdpStack.Port))
        self.assertEqual(device.kind, None)

        stack.close()

        ha = ('127.0.0.1', 8000)
        stack = stacking.UdpStack(ha=ha)
        self.assertIsInstance(stack.local, devicing.IpLocalDevice)
        self.assertEqual(stack.local.uid, 1)
        self.assertEqual(stack.local.name, "Device{0}".format(stack.local.uid))
        self.assertEqual(stack.local.ha, ha)
        self.assertEqual(stack.local.kind, None)
        self.assertEqual(stack.aha, ha)

        device = devicing.IpDevice(stack=stack, ha=ha)
        self.assertEqual(device.stack, stack)
        self.assertEqual(device.uid, 2)  # stacks .local has uid 1
        self.assertEqual(device.name, "Device{0}".format(device.uid))
        self.assertEqual(device.ha, ha)
        self.assertEqual(device.kind, None)
        self.assertEqual(device.host, ha[0])
        self.assertEqual(device.port, ha[1])

        device = devicing.IpRemoteDevice(stack=stack, ha=ha)
        self.assertIs(device.stack, stack)
        self.assertEqual(device.uid, 3)
        self.assertEqual(device.name, "Device{0}".format(device.uid))
        self.assertEqual(device.ha, ha)
        self.assertEqual(device.kind, None)
        self.assertEqual(device.host, ha[0])
        self.assertEqual(device.port, ha[1])

        stack.close()


def runOne(test):
    '''
    Unittest Runner
    '''
    test = BasicTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runSome():
    """ Unittest runner """
    tests =  []
    names = [
             'testDevice',
             'testUdpDevice',
            ]
    tests.extend(map(BasicTestCase, names))
    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)

def runAll():
    """ Unittest runner """
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(BasicTestCase))
    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:

    #runAll() #run all unittests

    runSome()#only run some

    #runOne('testPart')






"""
vectoring.py basic vector functions

"""
from __future__ import absolute_import, division, print_function

import sys
import math
from collections import namedtuple

# Import ioflo libs
from .sixing import *

from ..base.globaling import Pxy, Pxyz, Pne, Pned,  Pfs,  Pfsb  # namedtuple points

from .consoling import getConsole

console = getConsole()


def mag(v):
    """
    Returns the euclidean length or magnitude of vector v
    """
    return (pow(sum(e*e for e in v), 0.5))  # pow works for complex, math.sqrt does not

def mag2(v):
    """
    Returns the magnitude squared of vector v
    """
    return (sum(e*e for e in v))

def neg(v):
    """
    Returns the negative of v
    """
    return tuple(-e for e in v)

def mult(k, v):
    """
    Returns the vector result of multipling scalar k with vector v
    """
    return tuple(k*e for e in v)

def add(u, v):
    """
    Returns the vector addition of vectors u + v
    """
    if len(u) != len(v):
        raise ValueError("Vectors lengths differ {} != {}".format(len(u),
                                                                  len(v)))
    return tuple(e + g for e, g in zip(u, v))

def sub(u, v):
    """
    Returns the vector subtraction of vectors u - v
    """
    if len(u) != len(v):
        raise ValueError("Vectors lengths differ {} != {}".format(len(u),
                                                                  len(v)))
    return tuple(e - g for e, g in zip(u, v))

def dot(u, v):
    """
    Returns dot product of vectors u and v
    """
    if len(u) != len(v):
        raise ValueError("Vectors lengths differ {} != {}".format(len(u),
                                                                  len(v)))
    return (sum(e*g for e, g in zip(u, v)))

def norm(v, check=False):
    """
    Returns normalized (euclidean norm) copy of vector v as tuple
    If v is zero vector then returns v as is

    For non zero v If check is True then and if mag != 1.0 repeat normalization once
    """
    if not any(v):
        return v  # v is zero vector
    m = mag(v)
    nv = tuple(e / m for e in v )
    if check:
        if mag(nv) != 1.0:
            nv = norm(nv, check=check)
    return (nv)

def proj(u, v):
    """
    Returns the vector projection of u onto v, u proj v
    """
    w = norm(v)
    return (mult(dot(u, w), w))

def trip(u, v):
    """
    Returns the scalar triple product of vectors u and v and z axis.
    The convention is z dot (u cross v). Dotting with the z axis simplifies
    it to the z component of the u cross v
    The product is:
        positive if v is to the left of u, that is,
          the shortest right hand rotation from u to v is ccw
        negative if v is to the right of u, that is,
          the shortest right hand rotation from u to v is cw
        zero if v is colinear with u
    Essentially trip is the z component of the cross product of u x v
    """
    return (u[0] * v[1] - u[1] * v[0])

def ccw(u, v):
    """
    Returns True if v is to the left of u, that is,
          the shortest rotation from u to v is ccw, turn left
    False otherwise
    """
    return ((u[0] * v[1] - u[1] * v[0]) > 0)

left = ccw  # alias

def cw(u, v):
    """
    Returns True if v is to the right of u, that is,
          the shortest rotation from u to v is cw, turn right
    False otherwise
    """
    return ((u[0] * v[1] - u[1] * v[0]) < 0)

right = cw  #alias

def tween(p, u, v):
    """
    Returns true if point p is on the line from u to v, ie is between u and v
    """
    # treat u as the origin and create vectors a=up b=uv
    a = sub(p, u)
    b = sub(v, u)

    if a == b:  # endpoint same
        return True

    daa = dot(a, a)
    if daa == 0:  # p equals u so on line segment
        return True

    dbb = dot(b, b)
    if dbb == 0: # empty line segment
        if a == b:  # point on line only if same point
            return True
        else:  # point not on line
            return False

    dab = dot(a, b)
    if dab < 0:  # not on segment before
        return False

    if dab > dbb:  # not on segment after
        return False

    # collinear if ka=b
    # colinear is scalars all the same so any scale should work
    k = 0
    for e, g in zip(a, b):
        if e == 0:
            if g != 0:
                return False
            else:
                continue
        k = g / e
        break

    if k == 0:
        return False

    if mult(k, a) != b:
        return False

    return True

def tween2(p, u, v):
    """
    Returns true if 2d point p is on the line from 2d points u to v,
    ie is between u and v
    """
    # treat u as the origin and create vectors a=up b=uv
    a = sub(p, u)
    b = sub(v, u)

    dbb = dot(b, b)
    if dbb == 0: # empty line segment
        if a == b:  # point on line only if same point
            return True
        else:  # point not on line
            return False

    dab = dot(a, b)
    if dab < 0:
        return False  # before not between
    if dab > mag2(b):
        return False  # past not between
    if trip(a, b) != 0:  # not colinear
        return False
    return True

def cross3(u, v):
    """
    Returns 3 tuple that is  the 3 dimentional vector cross product of
    3 vector u crossed onto 3 vector v
    """
    return ((u[1] * v[2] - v[1] * u[2],
             u[2] * v[0] - v[2] * u[0],
             u[0] * v[1] - v[0] * u[1]))


def wind(p, vs):
    """
    Returns integer winding number of 2D point-in-polygon test for point p in
    the closed polygon whose sides are formed by connecting the vertex points
    in sequence vs where each element of vs is a point.
    Points are 2D sequences (tuples, namedtuples, lists etc)

    If the winding number is zero then p is not inside the polygon.
    If the winding number is positive then p is inside and the polygon is
      wound counter clockwise around p
    If the winding number is negative then p is inside and the polygon is
      wound clockwise around p



    Modified version of algorithm from Dan Sunday
    http://geomalgorithms.com/a03-_inclusion.html
    The Sunday algorithm does not handle the case consistently where p is on a vertex
    This modified algorithm accounts for p on edge (vertex or side) as outside
    Because P on the edge could be either inside or outside arbitrarily but there
    is no good way to determine winding direction in that case it is considered outside
    Algorithm checks for p as a vertex or side and returns w=0

    To determine which side of the infinite line directed from
    A=(x1,y1) to B=(x2,y2), a point P=(x,y) falls on
    compute the value:
    d=(x-x1)(y2-y1)-(y-y1)(x2-x1)
    u[0] * v[1] - u[1] * v[0]

    If d<0  then the point lies on one side of the line,
    If d>0 then it lies on the other side.
    If d=0 then the point lies exactly on the line.
    d is a version of the scalar triple product for the xy plane
    d is the z component of AB x AP

    """
    if p in vs:  # if on a vertex can't tell winding direction so outside
        return 0

    px, py = p[:2]
    w = 0  # winding number
    l = len(vs) # number of vertices
    for i in range(l):
        j = (i + 1) % l  # wrap around next vertex
        if tween2(p, vs[i], vs[j]):  # on side so outside
            return 0
        x, y = vs[i][:2]  # current vertex elements
        u, v = vs[j][:2]  # next vertex elements

        if y <= py:  # segment starts below ray level with x
            if v > py:  # segment ends above ray level with x = upward crossing
                # vertex vs[i] is origin, create vectors to p and vs[j]
                # compute if point is to the left of line segment by turn of vector
                if right(sub(p, vs[i]), sub(vs[j], vs[i])):  # turn right = is left
                    w += 1
        else:  # segment starts above ray level with x
            if v <= py: # segment ends below ray level with x = downward crossing
                # vertex vs[i] is origin, create vectors to p  and vs[j]
                # compute if point is to the right of line segment by turn of vector
                if left(sub(p, vs[i]), sub(vs[j], vs[i])):  # turn left = is right
                    w -= 1
    return (w)


def inside(p, vs, side=True):
    """
    Returns True if 2D point p is inside 2D polygon given by vs, False otherwise
    If side is True then p on an edge or vertex are considered inside.

    The closed polygon whose sides are formed by connecting the vertex points
    in sequence vs where each element of vs is a point.
    Points are 2D sequences (tuples, namedtuples, lists etc)

    If the winding number is zero then p is not inside the polygon.
    If the winding number is positive then p is inside and the polygon is
      wound counter clockwise around p
    If the winding number is negative then p is inside and the polygon is
      wound clockwise around p

    Modified version of algorithm from Dan Sunday
    http://geomalgorithms.com/a03-_inclusion.html
    The Sunday algorithm does not handle the case consistently where p is on a vertex
    This modified algorithm accounts for p on edge (vertex or side) as outside
    Because P on the edge could be either inside or outside arbitrarily but there
    is no good way to determine winding direction in that case it is considered outside
    Algorithm checks for p as a vertex or side and returns w=0

    To determine which side of the infinite line directed from
    A=(x1,y1) to B=(x2,y2), a point P=(x,y) falls on
    compute the value:
    d=(x-x1)(y2-y1)-(y-y1)(x2-x1)

    u[0] * v[1] - u[1] * v[0]

    If d<0  then the point lies on one side of the line,
    If d>0 then it lies on the other side.
    If d=0 then the point lies exactly on the line.
    d is a version of the scalar triple product for the xy plane
    d is the z component of AB x AP

    """
    if p in vs:  # on a vertex so use value of side
        return side

    px, py = p[:2]
    w = 0  # winding number
    l = len(vs) # number of vertices
    for i in range(l):
        j = (i + 1) % l  # wrap around next vertex
        if tween2(p, vs[i], vs[j]):  # on a side edge so use value of side
            return side
        x, y = vs[i][:2]  # current vertex elements
        u, v = vs[j][:2]  # next vertex elements

        if y <= py:  # segment starts below ray level with x
            if v > py:  # segment ends above ray level with x = upward crossing
                # vertex vs[i] is origin, create vectors to p and vs[j]
                # compute if point is to the left of line segment by turn of vector
                if right(sub(p, vs[i]), sub(vs[j], vs[i])):  # turn right = is left
                    w += 1
        else:  # segment starts above ray level with x
            if v <= py: # segment ends below ray level with x = downward crossing
                # vertex vs[i] is origin, create vectors to p  and vs[j]
                # compute if point is to the right of line segment by turn of vector
                if left(sub(p, vs[i]), sub(vs[j], vs[i])):  # turn left = is right
                    w -= 1

    return False if w == 0 else True  # w is inside if non-zero pos or neg


def insideOnly(p, vs):
    """
    Returns True if p is strictly inside polygon given by vs, False otherwise.
    If p on edge or vertex it is considered as outside
    """
    return inside(p, vs, side=False) # return False if wind(p, vs) == 0 else True


def outside(p, vs, side=True):
    """
    Returns True if p is outside polygon given by vs, False otherwise.
    If side is true then p on edge or vertex is considered as outside.
    """
    return False if inside(p, vs, side=not side) else True


def outsideOnly(p, vs):
    """
    Returns True if p is strictly outside polygon given by vs, False otherwise.
    If p on edge or vertex it is considered outside
    """
    return outside(p, vs, side=False)


def sideOnly(p, vs):
    """
    Returns True if 2D point p lies on an edge or vertex, False otherwise

    The closed polygon whose sides are formed by connecting the vertex points
    in sequence vs where each element of vs is a point.
    Points are 2D sequences (tuples, namedtuples, lists etc)

    """
    if p in vs:  # on a vertex
        return True
    px, py = p[:2]
    l = len(vs) # number of vertices
    for i in range(l):
        j = (i + 1) % l  # wrap around next vertex
        if tween2(p, vs[i], vs[j]):  # on a side edge
            return True
    return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Runs all the example FloScripts


"""
import sys
import os

import ioflo.app.run

PLAN_DIR_PATH = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    , 'plan')

def run(plan):
    """
    Run the  example script named plan
    """

    plan = os.path.join(PLAN_DIR_PATH, plan)
    plan = os.path.abspath(plan)
    root, ext = os.path.splitext(plan)
    if ext != '.flo' or root.startswith('__'):
        print("**** Invalid plan file = {0}.\n".format(plan))
        sys.exit()

    name, ext = os.path.splitext(os.path.basename(plan))
    skeddar = ioflo.app.run.run(  name=name,
                        filepath=plan,
                        period=0.0625,
                        verbose=2,
                        real=False,)

    print("Running Plan: {0}\n".format(plan))
    failed = False
    for house in skeddar.houses:
        failure = house.metas['failure'].value
        if failure:
            failed = True
            print("**** Failed in House = {0}. "
                  "Failure = {1}.\n".format(house.name, failure))
        else:
            print("**** Succeeded in House = {0}.\n".format(house.name))


if __name__ == '__main__':
    run('testRearRaze.flo')




"""
Device Base Package

"""
from __future__ import absolute_import, division, print_function


from ...aid import getConsole

console = getConsole()

class MixIn(object):
    """
    Base class to enable consistent MRO for mixin multiple inheritance
    """
    def __init__(self, *pa, **kwa):
        pass



# -*- coding: utf-8 -*-
"""
Unit Test Template
"""
from __future__ import absolute_import, division, print_function

import sys
import datetime
import unittest

import os

from ioflo.aid.sixing import *
from ioflo.aid.odicting import odict
from ioflo.test import testing
from ioflo.aid.consoling import getConsole
console = getConsole()

from ioflo.aid import byting


def setUpModule():
    console.reinit(verbosity=console.Wordage.concise)

def tearDownModule():
    pass


class BasicTestCase(unittest.TestCase):
    """
    Example TestCase
    """

    def setUp(self):
        """
        Call super if override so House Framer and Frame are setup correctly
        """
        super(BasicTestCase, self).setUp()

    def tearDown(self):
        """
        Call super if override so House Framer and Frame are torn down correctly
        """
        super(BasicTestCase, self).tearDown()

    def testBinizeUnbinize(self):
        """
        Test the binize unbinize functions
        """
        console.terse("{0}\n".format(self.testBinizeUnbinize.__doc__))

        n = 5
        u = byting.binize(n, 8)
        self.assertEqual(u, '00000101')
        n = byting.unbinize(u)
        self.assertEqual(n, 5)

    def testBytifyUnbytify(self):
        """
        Test the bytify unbytify functions
        """
        console.terse("{0}\n".format(self.testBytifyUnbytify.__doc__))

        b = bytearray([1, 2, 3])
        n = byting.unbytify(b)
        self.assertEqual(n, 0x010203)
        n = byting.unbytify([1, 2, 3])
        self.assertEqual(n, 0x010203)
        n = byting.unbytify(b'\x01\x02\x03')
        self.assertEqual(n, 0x010203)

        b = byting.bytify(n)
        self.assertEqual(b, bytearray([1, 2, 3]))
        b = byting.bytify(n, 4)
        self.assertEqual(b, bytearray([0, 1, 2, 3]))
        b = byting.bytify(n, 2)
        self.assertEqual(b, bytearray([1, 2, 3]))
        b = byting.bytify(n, 2, strict=True)
        self.assertEqual(b, bytearray([2, 3]))

        b = bytearray([1, 2, 3])
        n = byting.unbytify(b, reverse=True)
        self.assertEqual(n, 0x030201)
        n = byting.unbytify([1, 2, 3], reverse=True)
        self.assertEqual(n, 0x030201)
        n = byting.unbytify(b'\x01\x02\x03', reverse=True)
        self.assertEqual(n, 0x030201)

        b = byting.bytify(n, reverse=True)
        self.assertEqual(b, bytearray([1, 2, 3]))
        b = byting.bytify(n, 4, reverse=True)
        self.assertEqual(b, bytearray([1, 2, 3, 0]))
        b = byting.bytify(n, 2, reverse=True)
        self.assertEqual(b, bytearray([1, 2, 3]))
        b = byting.bytify(n, 2, reverse=True, strict=True)
        self.assertEqual(b, bytearray([1, 2]))

        n = -1
        b = byting.bytify(n)
        self.assertEqual(b, bytearray([0xff]))
        b = byting.bytify(n, 4)
        self.assertEqual(b, bytearray([0xff, 0xff, 0xff, 0xff]))
        n = -2
        b = byting.bytify(n)
        self.assertEqual(b, bytearray([0xfe]))
        b = byting.bytify(n, 2)
        self.assertEqual(b, bytearray([0xff, 0xfe]))

        n = 0x007ffb
        b = byting.bytify(n, 3)
        self.assertEqual(b, bytearray([0x00, 0x7f, 0xfb]))
        b = byting.bytify(n, 3, reverse=True)
        self.assertEqual(b, bytearray([0xfb, 0x7f, 0x00]))


    def testPackifyUnpackify(self):
        """
        Test the packbits
        """
        console.terse("{0}\n".format(self.testPackifyUnpackify.__doc__))

        fmt = u'3 2 1 1'
        fields = [6, 2, True, False]
        size = 1
        packed = byting.packify(fmt=fmt, fields=fields, size=size)
        self.assertEqual(packed, bytearray([0xd4]))
        self.assertEqual(byting.binize(byting.unbytify(packed), size*8), '11010100')

        packed = byting.packify(fmt=fmt, fields=fields)
        self.assertEqual(packed, bytearray([0xd4]))
        self.assertEqual(byting.binize(byting.unbytify(packed), size*8), '11010100')

        fmt = u''
        fields = []
        size = 0
        packed = byting.packify(fmt=fmt, fields=fields, size=size)
        self.assertEqual(packed, bytearray([]))

        packed = byting.packify(fmt=fmt, fields=fields)
        self.assertEqual(packed, bytearray([]))

        fmt = u'3 1'
        fields = [5, True]
        size = 1
        packed = byting.packify(fmt=fmt, fields=fields, size=size)
        self.assertEqual(packed, bytearray([0xb0]))
        self.assertEqual(byting.binize(byting.unbytify(packed), size*8), '10110000')

        fmt = u'8 6 7 3'
        fields = [0xA5, 0x38, 0x08, 0x01]
        size = 3
        packed = byting.packify(fmt=fmt, fields=fields, size=size)
        self.assertEqual(packed, bytearray([0xa5, 0xe0, 0x41]))
        self.assertEqual(byting.binize(byting.unbytify(packed), size*8), '101001011110000001000001')
        # 0xa5e040
        packed = byting.packify(fmt=fmt, fields=fields)
        self.assertEqual(packed, bytearray([0xa5, 0xe0, 0x41]))
        self.assertEqual(byting.binize(byting.unbytify(packed), size*8), '101001011110000001000001')

        fmt = u'8 6 7 3'
        fields = [0xA5, 0x38, 0x08, 0x01]
        size = 4
        packed = byting.packify(fmt=fmt, fields=fields, size=size)
        self.assertEqual(packed, bytearray([0xa5, 0xe0, 0x41, 0x00]))

        fmt = u'3 2 1 1'
        packed = bytearray([0xd4])
        fields = byting.unpackify(fmt=fmt, b=packed, boolean=True)
        self.assertEqual(fields, (6, 2, True, False, False))
        fields = byting.unpackify(fmt=fmt, b=packed, boolean=False)
        self.assertEqual(fields, (6, 2, 1, 0, 0))

        fmt = u''
        packed = bytearray([])
        fields = byting.unpackify(fmt=fmt, b=packed)
        self.assertEqual(fields, tuple())

        fmt = u'3 1'
        packed = bytearray([0xb0])
        fields = byting.unpackify(fmt=fmt, b=packed)
        self.assertEqual(fields, (5, 1, 0))

        fmt = u'4 3 1'
        packed = [0x0b]  # converts to bytearray in parameter
        fields = byting.unpackify(fmt=fmt, b=packed)
        self.assertEqual(fields, (0, 5, 1))

        fmt = u'8 6 7 3'
        packed = bytearray([0xa5, 0xe0, 0x41])
        fields = byting.unpackify(fmt=fmt, b=packed)
        self.assertEqual(fields, (0xA5, 0x38, 0x08, 0x01))

        fmt = u'8 6 7 3'
        packed = bytearray([0xa5, 0xe0, 0x41, 0xff, 0xff])
        fields = byting.unpackify(fmt=fmt, b=packed, size=3)
        self.assertEqual(fields, (0xA5, 0x38, 0x08, 0x01))

        fmt = u'3 2 1 1'
        fields = [6, 2, True, False]
        size = 1
        packed = byting.packify(fmt=fmt, fields=fields, size=size, reverse=True)
        self.assertEqual(packed, bytearray([0xd4]))
        self.assertEqual(byting.binize(byting.unbytify(packed), size*8), '11010100')

        packed = byting.packify(fmt=fmt, fields=fields, reverse=True)
        self.assertEqual(packed, bytearray([0xd4]))
        self.assertEqual(byting.binize(byting.unbytify(packed), size*8), '11010100')

        fmt = u''
        fields = []
        size = 0
        packed = byting.packify(fmt=fmt, fields=fields, size=size, reverse=True)
        self.assertEqual(packed, bytearray([]))

        packed = byting.packify(fmt=fmt, fields=fields, reverse=True)
        self.assertEqual(packed, bytearray([]))

        fmt = u'3 1'
        fields = [5, True]
        size = 1
        packed = byting.packify(fmt=fmt, fields=fields, size=size, reverse=True)
        self.assertEqual(packed, bytearray([0xb0]))
        self.assertEqual(byting.binize(byting.unbytify(packed), size*8), '10110000')

        fmt = u'8 6 7 3'
        fields = [0xA5, 0x38, 0x08, 0x01]
        size = 3
        packed = byting.packify(fmt=fmt, fields=fields, size=size, reverse=True)
        self.assertEqual(packed, bytearray([0x41, 0xe0, 0xa5]))
        self.assertEqual(byting.binize(byting.unbytify(packed), size*8), '010000011110000010100101')
        # 0xa5e040
        packed = byting.packify(fmt=fmt, fields=fields, reverse=True)
        self.assertEqual(packed, bytearray([0x41, 0xe0, 0xa5]))
        self.assertEqual(byting.binize(byting.unbytify(packed), size*8), '010000011110000010100101')

        fmt = u'8 6 7 3'
        fields = [0xA5, 0x38, 0x08, 0x01]
        size = 4
        packed = byting.packify(fmt=fmt, fields=fields, size=size, reverse=True)
        self.assertEqual(packed, bytearray([0x00, 0x41, 0xe0, 0xa5]))

        fmt = u'3 2 1 1'
        packed = bytearray([212])
        fields = byting.unpackify(fmt=fmt, b=packed, boolean=True, reverse=True)
        self.assertEqual(fields, (6, 2, True, False, False))
        fields = byting.unpackify(fmt=fmt, b=packed, boolean=False, reverse=True)
        self.assertEqual(fields, (6, 2, 1, 0, 0))

        fmt = u''
        packed = bytearray([])
        fields = byting.unpackify(fmt=fmt, b=packed, reverse=True)
        self.assertEqual(fields, tuple())

        fmt = u'3 1'
        packed = bytearray([0xb0])
        fields = byting.unpackify(fmt=fmt, b=packed, reverse=True)
        self.assertEqual(fields, (5, 1, 0))

        fmt = u'4 3 1'
        packed = [0x0b]  # converts to bytearray in parameter
        fields = byting.unpackify(fmt=fmt, b=packed, reverse=True)
        self.assertEqual(fields, (0, 5, 1))

        fmt = u'8 6 7 3'
        packed = bytearray([0xa5, 0xe0, 0x41])
        fields = byting.unpackify(fmt=fmt, b=packed, reverse=True)
        self.assertEqual(fields, (0x41, 0x38, 0x14, 0x05))

        fmt = u'8 6 7 3'
        packed = bytearray([0xff, 0xff, 0xa5, 0xe0, 0x41])
        fields = byting.unpackify(fmt=fmt, b=packed, size=3, reverse=True)
        self.assertEqual(fields, (0x41, 0x38, 0x14, 0x05))


    def testPackifyInto(self):
        """
        Test the packbits
        """
        console.terse("{0}\n".format(self.testPackifyInto.__doc__))

        fmt = u'3 2 1 1'
        fields = [6, 2, True, False]
        b = bytearray([0, 0, 0, 0, 0, 0])
        size = byting.packifyInto(b, fmt=fmt, fields=fields, size=1)
        self.assertEqual(size, 1)
        self.assertEqual(b, bytearray([212, 0, 0, 0, 0, 0]))
        self.assertEqual(byting.binize(b[0]), '11010100')

        size = byting.packifyInto(b, fmt=fmt, fields=fields, size=1, offset=2)
        self.assertEqual(size, 1)
        self.assertEqual(b, bytearray([212, 0, 212, 0, 0, 0]))
        self.assertEqual(byting.binize(b[2]), '11010100')

        b[0] =  0x0
        self.assertEqual(b, bytearray([0, 0, 212, 0, 0, 0]))
        size = byting.packifyInto(b, fmt=fmt, fields=fields)
        self.assertEqual(size, 1)
        self.assertEqual(b, bytearray([212, 0, 212, 0, 0, 0]))
        self.assertEqual(byting.binize(b[0]), '11010100')

        b = bytearray([0, 0, 0, 0, 0, 0])
        fmt = u''
        fields = []
        size = byting.packifyInto(b, fmt=fmt, fields=fields, size=0)
        self.assertEqual(size, 0)
        self.assertEqual(b, bytearray([0, 0, 0, 0, 0, 0]))

        size = byting.packifyInto(b, fmt=fmt, fields=fields)
        self.assertEqual(size, 0)
        self.assertEqual(b, bytearray([0, 0, 0, 0, 0, 0]))

        fmt = u'3 1'
        fields = [5, True]
        size = byting.packifyInto(b, fmt=fmt, fields=fields, size=1)
        self.assertEqual(size, 1)
        self.assertEqual(b, bytearray([176, 0, 0, 0, 0, 0]))
        self.assertEqual(byting.binize(b[0]), '10110000')

        b = bytearray([0, 0, 0, 0, 0, 0])
        fmt = u'8 6 7 3'
        fields = [0xA5, 0x38, 0x08, 0x01]
        size = byting.packifyInto(b, fmt=fmt, fields=fields)
        self.assertEqual(size, 3)
        self.assertEqual(b, bytearray([0xa5, 0xe0, 0x41, 0, 0, 0]))

        b = bytearray([0, 0, 0, 0, 0, 0])
        size = byting.packifyInto(b, fmt=fmt, fields=fields, offset=1)
        self.assertEqual(size, 3)
        self.assertEqual(b, bytearray([0, 0xa5, 0xe0, 0x41, 0, 0]))

        b = bytearray()
        fmt = u'8 6 7 3'
        fields = [0xA5, 0x38, 0x08, 0x01]
        size = byting.packifyInto(b, fmt=fmt, fields=fields)
        self.assertEqual(size, 3)
        self.assertEqual(b, bytearray([0xa5, 0xe0, 0x41]))

        fmt = u'3 2 1 1'
        fields = [6, 2, True, False]
        b = bytearray([0, 0, 0, 0, 0, 0])
        size = byting.packifyInto(b, fmt=fmt, fields=fields, size=1, reverse=True)
        self.assertEqual(size, 1)
        self.assertEqual(b, bytearray([212, 0, 0, 0, 0, 0]))
        self.assertEqual(byting.binize(b[0]), '11010100')

        size = byting.packifyInto(b, fmt=fmt, fields=fields, size=1, offset=2, reverse=True)
        self.assertEqual(size, 1)
        self.assertEqual(b, bytearray([212, 0, 212, 0, 0, 0]))
        self.assertEqual(byting.binize(b[2]), '11010100')

        b[0] =  0x0
        self.assertEqual(b, bytearray([0, 0, 212, 0, 0, 0]))
        size = byting.packifyInto(b, fmt=fmt, fields=fields, reverse=True)
        self.assertEqual(size, 1)
        self.assertEqual(b, bytearray([212, 0, 212, 0, 0, 0]))
        self.assertEqual(byting.binize(b[0]), '11010100')

        b = bytearray([0, 0, 0, 0, 0, 0])
        fmt = u''
        fields = []
        size = byting.packifyInto(b, fmt=fmt, fields=fields, size=0, reverse=True)
        self.assertEqual(size, 0)
        self.assertEqual(b, bytearray([0, 0, 0, 0, 0, 0]))

        size = byting.packifyInto(b, fmt=fmt, fields=fields, reverse=True)
        self.assertEqual(size, 0)
        self.assertEqual(b, bytearray([0, 0, 0, 0, 0, 0]))

        fmt = u'3 1'
        fields = [5, True]
        size = byting.packifyInto(b, fmt=fmt, fields=fields, size=1, reverse=True)
        self.assertEqual(size, 1)
        self.assertEqual(b, bytearray([176, 0, 0, 0, 0, 0]))
        self.assertEqual(byting.binize(b[0]), '10110000')

        b = bytearray([0, 0, 0, 0, 0, 0])
        fmt = u'8 6 7 3'
        fields = [0xA5, 0x38, 0x08, 0x01]
        size = byting.packifyInto(b, fmt=fmt, fields=fields, reverse=True)
        self.assertEqual(size, 3)
        self.assertEqual(b, bytearray([0x41, 0xe0, 0xa5, 0, 0, 0]))

        b = bytearray([0, 0, 0, 0, 0, 0])
        size = byting.packifyInto(b, fmt=fmt, fields=fields, offset=1, reverse=True)
        self.assertEqual(size, 3)
        self.assertEqual(b, bytearray([0, 0x41, 0xe0, 0xa5, 0, 0]))

        b = bytearray()
        fmt = u'8 6 7 3'
        fields = [0xA5, 0x38, 0x08, 0x01]
        size = byting.packifyInto(b, fmt=fmt, fields=fields, reverse=True)
        self.assertEqual(size, 3)
        self.assertEqual(b, bytearray([0x41, 0xe0, 0xa5]))

    def testSignExtend(self):
        """
        Test the signExtend function
        """
        console.terse("{0}\n".format(self.testSignExtend.__doc__))
        a =( 2 ** 8) - 11  # 8 bit two's complement rep of -11
        self.assertEqual(a, 0xf5)
        b = bytearray([0xf5])  # make bytearray
        y, x = byting.unpackify(fmt='3 5', b=b)  # assign lower 5 bits to x
        self.assertEqual(x, 0x15)
        z = byting.signExtend(x, n=5)
        self.assertEqual(z, -11)

        a = 11
        self.assertEqual(a, 0x0b)
        b = bytearray([0x0b])
        y, x = byting.unpackify(fmt='3 5', b=b)  # assign lower 5 bits to x
        self.assertEqual(x, 0x0b)
        z = byting.signExtend(x, n=5)
        self.assertEqual(z, 11)

        a = (2 ** 8) - 16
        self.assertEqual(a, 0xf0)
        b = bytearray([0xf0])  # make bytearray
        y, x = byting.unpackify(fmt='3 5', b=b)  # assign lower 5 bits to x
        self.assertEqual(x, 0x10)
        z = byting.signExtend(x, n=5)
        self.assertEqual(z, -16)

        a = 15
        self.assertEqual(a, 0x0f)
        b = bytearray([0x0f])
        y, x = byting.unpackify(fmt='3 5', b=b)  # assign lower 5 bits to x
        self.assertEqual(x, 0x0f)
        z = byting.signExtend(x, n=5)
        self.assertEqual(z, 15)

        a = 0
        self.assertEqual(a, 0x00)
        b = bytearray([0x00])
        y, x = byting.unpackify(fmt='3 5', b=b)  # assign lower 5 bits to x
        self.assertEqual(x, 0x00)
        z = byting.signExtend(x, n=5)
        self.assertEqual(z, 0)



def runOne(test):
    '''
    Unittest Runner
    '''
    test = BasicTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runSome():
    """ Unittest runner """
    tests =  []
    names = [
             'testBinizeUnbinize',
             'testBytifyUnbytify',
             'testPackifyUnpackify',
             'testPackifyInto',
             'testSignExtend',
            ]
    tests.extend(map(BasicTestCase, names))
    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)

def runAll():
    """ Unittest runner """
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(BasicTestCase))
    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:

    #console.reinit(verbosity=console.Wordage.concise)

    #runAll() #run all unittests

    runSome()#only run some

    #runOne('testBasic')





# -*- coding: utf-8 -*-
"""
Unit Test Template
"""

import sys
import unittest

import os
from ioflo.aid import odict
from ioflo.test import testing
from ioflo.aid.consoling import getConsole
console = getConsole()

from ioflo.aid import classing
from ioflo.base import registering

def setUpModule():
    console.reinit(verbosity=console.Wordage.concise)

def tearDownModule():
    pass


class BasicTestCase(unittest.TestCase):
    """
    Example TestCase
    """

    def setUp(self):
        """
        Call super if override so House Framer and Frame are setup correctly
        """
        super(BasicTestCase, self).setUp()

    def tearDown(self):
        """
        Call super if override so House Framer and Frame are torn down correctly
        """
        super(BasicTestCase, self).tearDown()

    def testNonStringIterable(self):
        """
        Test the utility function nonStringIterable
        """
        console.terse("{0}\n".format(self.testNonStringIterable.__doc__))
        a = bytearray(b'abc')
        w = dict(a=1, b=2, c=3)
        x = 'abc'
        y = b'abc'
        z = [1, 2, 3]

        self.assertTrue(isinstance(a, classing.NonStringIterable))
        self.assertTrue(isinstance(w, classing.NonStringIterable))
        self.assertFalse(isinstance(x, classing.NonStringIterable))
        self.assertFalse(isinstance(y, classing.NonStringIterable))
        self.assertTrue(isinstance(z, classing.NonStringIterable))


    def testNonStringSequence(self):
        """
        Test the utility function nonStringSequence
        """
        console.terse("{0}\n".format(self.testNonStringSequence.__doc__))
        a = bytearray(b'abc')
        w = dict(a=1, b=2, c=3)
        x = 'abc'
        y = b'abc'
        z = [1, 2, 3]

        self.assertTrue(isinstance(a, classing.NonStringIterable))
        self.assertFalse(isinstance(w, classing.NonStringSequence))
        self.assertFalse(isinstance(x, classing.NonStringSequence))
        self.assertFalse(isinstance(y, classing.NonStringSequence))
        self.assertTrue(isinstance(z, classing.NonStringSequence))

    def testMetaclassify(self):
        """
        Test the utility decorator metaclassify
        """
        console.terse("{0}\n".format(self.testMetaclassify.__doc__))

        @classing.metaclassify(registering.RegisterType)
        class A(object):
            #__metaclass__ = registering.RegisterType
            def __init__(self, name="", store=None):
                self.name = name
                self.store = store


        self.assertEqual(A.Registry, {'A': (A, None, None, None)})

    def testIsIterator(self):
        """
        Test the utility function isIterator
        """
        o = [1, 2, 3]
        self.assertFalse(classing.isIterator(o))
        i = iter(o)
        self.assertTrue(classing.isIterator(i))

        def genf():
            yield ""
            yield ""

        self.assertFalse(classing.isIterator(genf))
        gen = genf()
        self.assertTrue(classing.isIterator(gen))


    def testAttributize(self):
        """
        Test the utility decorator attributize generator
        """
        global headed

        console.terse("{0}\n".format(self.testAttributize.__doc__))

        def gf(skin, x):
            skin.x = 5
            skin.y = 'a'
            cnt = 0
            while cnt < x:
                yield cnt
                cnt += 1

        agf = classing.attributize(gf)
        ag = agf(3)
        self.assertTrue(classing.isIterator(ag))
        self.assertFalse(hasattr(ag, 'x'))
        self.assertFalse(hasattr(ag, 'y'))
        n = next(ag)  # body of gf is not run until first next call
        self.assertEqual(n, 0)
        self.assertTrue(hasattr(ag, 'x'))
        self.assertTrue(hasattr(ag, 'y'))
        self.assertEqual(ag.x, 5)
        self.assertEqual(ag.y, 'a')
        n = next(ag)
        self.assertEqual(n, 1)


        # Set up like WSGI for generator function
        @classing.attributize
        def bar(skin, req=None, rep=None):
            """
            Generator function with "skin" parameter for skin wrapper to
            attach attributes
            """
            skin._status = 400
            skin._headers = odict(example="Hi")
            yield b""
            yield b""
            yield b"Hello There"
            return b"Goodbye"

        # now use it like WSGI server does
        msgs = []
        gen = bar()
        self.assertTrue(classing.isIterator(gen))
        self.assertFalse(hasattr(gen, '_status'))
        self.assertFalse(hasattr(gen, '_headers'))

        headed = False

        def write(msg):
            """
            Simulate WSGI write
            """
            global headed

            if not headed:  # add headers
                if hasattr(gen, "_status"):
                    if gen._status is not None:
                        msgs.append(str(gen._status))
                if hasattr(gen, "_headers"):
                    for key, val in gen._headers.items():
                        msgs.append("{}={}".format(key, val))

                headed = True

            msgs.append(msg)

        igen = iter(gen)
        self.assertIs(igen, gen)
        done = False
        while not done:
            try:
                msg = next(igen)
            except StopIteration as ex:
                if hasattr(ex, "value") and ex.value:
                    write(ex.value)
                write(b'')  # in case chunked send empty chunk to terminate
                done = True
            else:
                if msg:  # only write if not empty allows async processing
                    write(msg)

        self.assertEqual(msgs, ['400', 'example=Hi', b'Hello There', b'Goodbye', b''])


        # Set up like WSGI for generator method
        class R:
            @classing.attributize
            def bar(self, skin, req=None, rep=None):
                """
                Generator function with "skin" parameter for skin wrapper to
                attach attributes
                """
                self.name = "Peter"
                skin._status = 400
                skin._headers = odict(example="Hi")
                yield b""
                yield b""
                yield b"Hello There " + self.name.encode()
                return b"Goodbye"

        # now use it like WSGI server does
        r = R()
        msgs = []
        gen = r.bar()
        self.assertTrue(classing.isIterator(gen))
        self.assertFalse(hasattr(gen, '_status'))
        self.assertFalse(hasattr(gen, '_headers'))

        headed = False

        def write(msg):
            """
            Simulate WSGI write
            """
            global headed

            if not headed:  # add headers
                if hasattr(gen, "_status"):
                    if gen._status is not None:
                        msgs.append(str(gen._status))
                if hasattr(gen, "_headers"):
                    for key, val in gen._headers.items():
                        msgs.append("{}={}".format(key, val))

                headed = True

            msgs.append(msg)

        igen = iter(gen)
        self.assertIs(igen, gen)
        done = False
        while not done:
            try:
                msg = next(igen)
            except StopIteration as ex:
                if hasattr(ex, "value") and ex.value:
                    write(ex.value)
                write(b'')  # in case chunked send empty chunk to terminate
                done = True
            else:
                if msg:  # only write if not empty allows async processing
                    write(msg)

        self.assertEqual(msgs, ['400', 'example=Hi', b'Hello There Peter', b'Goodbye', b''])


def runOne(test):
    '''
    Unittest Runner
    '''
    test = BasicTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runSome():
    """ Unittest runner """
    tests =  []
    names = ['testNonStringIterable',
             'testNonStringSequence',
             'testMetaclassify',
             'testIsIterator',
             'testAttributize',
            ]
    tests.extend(map(BasicTestCase, names))
    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)

def runAll():
    """ Unittest runner """
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(BasicTestCase))
    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:

    #console.reinit(verbosity=console.Wordage.concise)

    #runAll() #run all unittests

    runSome()#only run some

    #runOne('testBasic')
    #runOne('testSkinGenFunc')





# -*- coding: utf-8 -*-
"""
Unittests for udp async io (nonblocking) module
"""

import sys
import unittest

import os
import time
import tempfile
import shutil
import socket

from ioflo.aid.sixing import *
from ioflo.aid.consoling import getConsole
from ioflo.aio import wiring
from ioflo.aio.udp import udping
from ioflo.aio import aioing

console = getConsole()


def setUpModule():
    console.reinit(verbosity=console.Wordage.concise)

def tearDownModule():
    console.reinit(verbosity=console.Wordage.concise)

class BasicTestCase(unittest.TestCase):
    """
    Test Case
    """

    def setUp(self):
        """

        """
        pass

    def tearDown(self):
        """

        """
        pass

    def testSocketUdpNb(self):
        """
        Test Class SocketUdpNb
        """
        console.terse("{0}\n".format(self.testSocketUdpNb.__doc__))
        console.reinit(verbosity=console.Wordage.profuse)

        userDirpath = os.path.join('~', '.ioflo', 'test')
        userDirpath = os.path.abspath(os.path.expanduser(userDirpath))
        if not os.path.exists(userDirpath):
            os.makedirs(userDirpath)

        tempDirpath = tempfile.mkdtemp(prefix="test", suffix="log", dir=userDirpath)

        logDirpath = os.path.join(tempDirpath, 'log')
        if not os.path.exists(logDirpath):
            os.makedirs(logDirpath)

        wireLog = wiring.WireLog(path=logDirpath)
        result = wireLog.reopen(prefix='alpha', midfix='6101')

        alpha = udping.SocketUdpNb(port = 6101, wlog=wireLog)
        self.assertIs(alpha.opened, False)
        self.assertIs(alpha.reopen(), True)
        self.assertIs(alpha.opened, True)

        beta = udping.SocketUdpNb(port = 6102)
        self.assertIs(beta.opened, False)
        self.assertIs(beta.reopen(), True)
        self.assertIs(beta.opened, True)


        console.terse("Sending alpha to beta\n")
        msgOut = b"alpha sends to beta"
        alpha.send(msgOut, beta.ha)
        time.sleep(0.05)
        msgIn, src = beta.receive()
        self.assertEqual(msgOut, msgIn)
        self.assertEqual(src[1], alpha.ha[1])

        console.terse("Sending alpha to alpha\n")
        msgOut = b"alpha sends to alpha"
        alpha.send(msgOut, alpha.ha)
        time.sleep(0.05)
        msgIn, src = alpha.receive()
        self.assertEqual(msgOut, msgIn)
        self.assertEqual(src[1], alpha.ha[1])

        console.terse("Sending beta to alpha\n")
        msgOut = b"beta sends to alpha"
        beta.send(msgOut, alpha.ha)
        time.sleep(0.05)
        msgIn, src = alpha.receive()
        self.assertEqual(msgOut, msgIn)
        self.assertEqual(src[1], beta.ha[1])
        time.sleep(0.05)

        console.terse("Sending beta to beta\n")
        msgOut = b"beta sends to beta"
        beta.send(msgOut, beta.ha)
        time.sleep(0.05)
        time.sleep(0.05)
        msgIn, src = beta.receive()
        self.assertEqual(msgOut, msgIn)
        self.assertEqual(src[1], beta.ha[1])

        alpha.close()
        beta.close()

        self.assertIs(alpha.opened, False)

        wireLog.close()
        shutil.rmtree(tempDirpath)
        console.reinit(verbosity=console.Wordage.concise)

    def testBroadcast(self):
        """
        Test Class SocketUdpNb
        """
        console.terse("{0}\n".format(self.testBroadcast.__doc__))


        try:  # only run if netifaces installed
            from ioflo.aio.aioing import getDefaultHost,  getDefaultBroadcast
        except ImportError:
            return


        console.reinit(verbosity=console.Wordage.profuse)


        #unicast = socket.gethostbyname(socket.gethostname())
        unicast = getDefaultHost()

        #parts = unicast.split('.')
        #parts[3] = '255'
        #broadcast = ".".join(parts)  # make broadcast send address

        broadcast = getDefaultBroadcast()
        bha = (broadcast, 6102)

        alpha = udping.SocketUdpNb(host=unicast,
                                   port = 6101,
                                   bcast=True)
        self.assertIs(alpha.opened, False)
        self.assertIs(alpha.reopen(), True)
        self.assertIs(alpha.opened, True)
        self.assertIs(alpha.bcast, True)
        self.assertEqual(alpha.ha, (unicast, 6101))

        beta = udping.SocketUdpNb(host="",  # any
                                  port=6102)
        self.assertIs(beta.reopen(), True)
        self.assertIs(beta.bcast, False)
        self.assertEqual(beta.ha, ("0.0.0.0", 6102))
        self.assertEqual(beta.ha[1], bha[1])  # same port

        console.terse("Broadcasting alpha to beta\n")
        msgOut = b"alpha broadcasts to beta"
        alpha.send(msgOut, bha)
        time.sleep(0.1)
        msgIn, src = beta.receive()
        self.assertEqual(msgIn, msgOut)
        self.assertEqual(src, alpha.ha)

        alpha.close()
        beta.close()
        time.sleep(0.1)

        beta = udping.SocketUdpNb(host="",  # any
                                  port=6102,
                                  bcast=True)
        self.assertIs(beta.reopen(), True)
        self.assertIs(beta.bcast, True)
        self.assertEqual(beta.ha, ("0.0.0.0", 6102))
        self.assertEqual(beta.ha[1], bha[1])  # same port

        console.terse("Broadcasting beta to beta\n")
        msgOut = b"beta broadcasts to beta"
        beta.send(msgOut, bha)
        time.sleep(0.2)
        msgIn, src = beta.receive()
        self.assertEqual(msgIn, msgOut)
        self.assertEqual(src[1], beta.ha[1])

        beta.close()
        time.sleep(0.5)
        self.assertIs(alpha.opened, False)
        self.assertIs(beta.opened, False)
        console.reinit(verbosity=console.Wordage.concise)


def runOne(test):
    '''
    Unittest Runner
    '''
    test = BasicTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runSome():
    """ Unittest runner """
    tests =  []
    names = [
             'testSocketUdpNb',
             'testBroadcast',
            ]
    tests.extend(map(BasicTestCase, names))
    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)

def runAll():
    """ Unittest runner """
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(BasicTestCase))
    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:

    #console.reinit(verbosity=console.Wordage.concise)

    #runAll() #run all unittests

    runSome()#only run some

    #runOne('testClientAutoReconnect')





"""
Exchange Base Class Package

"""
from __future__ import absolute_import, division, print_function

import struct
from binascii import hexlify
from collections import deque
import enum

from ...aid.sixing import *
from ...aid.odicting import odict
from ...aid.byting import bytify, unbytify, packify, packifyInto, unpackify
from ...aid.eventing import eventify, tagify
from ...aid.timing import StoreTimer, tuuid
from ...aid import getConsole
from .protoing import MixIn

console = getConsole()


class Exchange(MixIn):
    """
    Exchange (pseudo transaction) base class for exchanges
    """
    Timeout =  2.0  # default timeout
    RedoTimeout = 0.5  # default redo timeout

    def __init__(self,
                 stack,
                 uid=None,
                 name=None,
                 device=None,
                 timeout=None,
                 redoTimout=None,
                 tx=None,
                 rx=None):
        """
        Setup Exchange instance

        Inherited Parameters:

        Parameters:
            stack is interface stack instance
            uid is Exchange unique id
            name is user friendly name of exchange
            device is associated device handling exchange
            timeout is exchange expiration timeout
                timeout of 0.0 means no expiration go on forever
            redoTimeout is redo appropriate packet/message in exchange
            rx is latest received  msg/pkt/data
            tx is latest/next transmitted msg/pkt/data

        Class Attributes:
            .Timeout is overall exchange timeout
            .RedoTimeout is redo timeout

        Inherited Attributes

        Attributes:
            .stack is interface stack instance
            .uid is Exchange unique id
            .name is user friendly name of exchange
            .device is associated device handling exchange
            .timeout is exchange expiration timeout
                timeout of 0.0 means no expiration go on forever
            .timer is StoreTimer instance for .timeout
            .redoTimeout is redo appropriate packet/message in exchange
            .redoTimer is StoreTimer instance for .redoTimeout
            .rx is latest received  msg/pkt/data
            .tx is latest/next transmitted msg/pkt/data
            .done is True If done  False otherwise
            .failed is True If failed False otherwise
            .acked is True if ack has been sent

        Inherited Properties

        Properties

        """
        self.stack = stack
        self.uid = uid if uid is not None else tuuid()
        self.name = name or self.__class__.__name__
        self.device = device
        self.timeout = timeout if timeout is not None else self.Timeout
        self.timer = StoreTimer(stack.stamper, duration=self.timeout)
        self.redoTimeout = redoTimeout if redoTimout is not None else self.RedoTimeout
        self.redoTimer = StoreTimer(stack.stamper, duration=self.redoTimeout)
        self.rx = rx  # latest received
        self.tx = tx  # initial to transmit
        self.done = False
        self.failed = False
        self.acked = False

    def prepStart(self):
        """
        Set flags for start
        """
        self.done = False
        self.failed = False
        self.acked = False

    def start(self):
        """
        Startup first run when context is ready
        override in subclass
        """
        self.prepStart()

    def run(self):
        """
        Run one iteration
        """
        self.done = True

    def process(self):
        """
        Process time based handling of exchange like timeout or retries
        """
        if self.timeout > 0.0 and self.timer.expired:
            console.verbose("{0}. Timed out with {1} at {2}\n".format(
                    self.stack.name, self.device.name, round(self.stack.stamper.stamp, 3)))
            self.fail()
            return

        if self.redoTimeout > 0.0 and self.redoTimer.expired:
            self.redoTimer.restart()
            console.verbose("{0}: Redoing {1} with {2} at {3}\n".format(
                                            self.stack.name,
                                            self.name,
                                            self.device.name,
                                            round(self.stack.stamper.stamp, 3)))
            if self.tx is not None:
                self.send(self.tx)

    def prepSend(self, tx=None):
        """
        Setups flags and .tx for send.
        """
        if tx is not None:
            self.tx = tx  # always last transmitted msg/pkt/data
        if self.tx is None:
            raise ValueError("{0}: {1} Cannot send. No .tx."
                                "\n".format(self.stack.name,
                                            self.name))

    def send(self, tx=None):
        """
        Setup and transmit packet

        Extend in subclass to queue appropriately

        Either:
           self.prepSend(tx)
           self.stack.transmit(self.tx, self.device.ha)
        Or
           self.prepSend(tx)
           self.stack.message(self.tx, self.device)
        """
        self.prepSend(tx=tx)
        self.transmit(pkt=tx)

    def transmit(self, pkt=None):
        """
        Queue pkt on stack packet queue
        """
        if pkt is not None:
            self.tx = pkt  # always last transmitted msg/pkt/data
        if self.tx is None:
            raise ValueError("{0}: {1} Cannot transmit. No .tx."
                                "\n".format(self.stack.name,
                                            self.name))
        self.stack.transmit(self.tx)

    def message(self, msg=None):
        """
        Queue msg on stack message queue
        """
        if msg is not None:
            self.tx = msg  # always last transmitted msg/pkt/data
        if self.tx is None:
            raise ValueError("{0}: {1} Cannot message. No .tx."
                                "\n".format(self.stack.name,
                                            self.name))
        self.stack.message(self.tx)

    def prepReceive(self, rx):
        """
        Setups .rx for receive. And flags if any
        extend in subclass
        """
        self.rx = rx

    def receive(self, rx):
        """
        Process received msg/pkt/data.
        Subclasses should extend with call to .prepReceive
        and additional handling
        """
        self.prepReceive(rx)


    def prepFinish(self):
        """
        Mark flags
        """
        self.done = True

    def finish(self):
        """
        Exchange complete
        Default is as success unless .failed
        Mark flags
        """
        self.prepFinish()
        console.verbose("{0}: Finished {1} with {2} as {3}"
                        " at {4}\n".format(self.stack.name,
                                            self.name,
                                            self.device.name,
                                            'FAILURE' if self.failed else 'SUCCESS',
                                            round(self.stack.stamper.stamp, 3)))

    def fail(self):
        """
        Exchange complete as failure
        """
        self.failed = True
        self.finish()


class Exchanger(Exchange):
    """
    Exchanger (pseudo transaction) base class for initiated exchanges
    """
    def __init__(self, stack, **kwa):
        """
        Setup Exchanger instance

        Inherited Parameters:
            stack is interface stack instance
            uid is Exchange unique id
            name is user friendly name of exchange
            device is associated device handling exchange
            timeout is exchange expiration timeout
                timeout of 0.0 means no expiration go on forever
            redoTimeout is redo appropriate packet/message in exchange
            rx is latest received  msg/pkt/data
            tx is latest/next transmitted msg/pkt/data

        Parameters:

        Inherited Class Attributes:
            .Timeout is overall exchange timeout
            .RedoTimeout is redo timeout

        Inherited Attributes
            .stack is interface stack instance
            .uid is Exchange unique id
            .name is user friendly name of exchange
            .device is associated device handling exchange
            .timeout is exchange expiration timeout
                timeout of 0.0 means no expiration go on forever
            .timer is StoreTimer instance for .timeout
            .redoTimeout is redo appropriate packet/message in exchange
            .redoTimer is StoreTimer instance for .redoTimeout
            .rx is latest received  msg/pkt/data
            .tx is latest/next transmitted msg/pkt/data
            .done is True If done  False otherwise
            .failed is True If failed False otherwise
            .acked is True if ack has been sent

        Attributes:

        Inherited Properties

        Properties

        """
        super(Exchanger, self).__init__(stack=stack, **kwa)

    def start(self, tx=None):
        """
        Initiate exchange
        tx is latest/next transmitted msg/pkt/data

        """
        self.prepStart()  # reset flags

        self.timer.restart()
        self.redoTimer.restart()
        console.verbose("{0}: Initiating {1} with {2} at {3}.\n".format(self.stack.name,
                                            self.name,
                                            self.device.name,
                                            round(self.stack.stamper.stamp, 3)))
        self.send(tx)

class Exchangent(Exchange):
    """
    Exchangent (pseudo transaction) base class for correspondent exchanges
    """
    Timeout = 0.5  # default timeout
    RedoTimeout = 0.1

    def __init__(self, stack, **kwa):
        """
        Setup Exchangent instance

        Inherited Parameters:
            stack is interface stack instance
            uid is Exchange unique id
            name is user friendly name of exchange
            device is associated device handling exchange
            timeout is exchange expiration timeout
                timeout of 0.0 means no expiration go on forever
            redoTimeout is redo appropriate packet/message in exchange
            rx is latest received  msg/pkt/data
            tx is latest/next transmitted msg/pkt/data

        Parameters:

        Inherited Class Attributes:
            .Timeout is overall exchange timeout
            .RedoTimeout is redo timeout

        Inherited Attributes
            .stack is interface stack instance
            .uid is Exchange unique id
            .name is user friendly name of exchange
            .device is associated device handling exchange
            .timeout is exchange expiration timeout
                timeout of 0.0 means no expiration go on forever
            .timer is StoreTimer instance for .timeout
            .redoTimeout is redo appropriate packet/message in exchange
            .redoTimer is StoreTimer instance for .redoTimeout
            .rx is latest received  msg/pkt/data
            .tx is latest/next transmitted msg/pkt/data
            .done is True If done  False otherwise
            .failed is True If failed False otherwise
            .acked is True if ack has been sent

        Attributes:

        Inherited Properties

        Properties


        """
        super(Exchangent, self).__init__(stack=stack, **kwa)

    def start(self, rx=None):
        """
        Correspond to exchange
        """
        self.prepStart()  # reset flags

        self.timer.restart()
        self.redoTimer.restart()
        console.verbose("{0}: Corresponding {1} with {2} at {3}.\n".format(self.stack.name,
                                            self.name,
                                            self.device.name,
                                            round(self.stack.stamper.stamp, 3)))
        self.respond(rx)

    def respond(self, rx=None):
        """
        Respond to the associated request
        """
        if rx is not None:
            self.rx = rx

        if self.rx is None:
            raise ValueError("{0}: Cannot respond {1} no .rx"
                             "received.\n".format(self.stack.name,
                                                  self.name))
        self.finish()



# -*- coding: utf-8 -*-
"""
Unittests for wiring (wire logging) module
Over the wire logs
"""

import sys
import unittest

import os
import tempfile
import shutil

from ioflo.aid.sixing import *
from ioflo.aid.consoling import getConsole
from ioflo.aio import wiring

console = getConsole()


def setUpModule():
    console.reinit(verbosity=console.Wordage.concise)

def tearDownModule():
    console.reinit(verbosity=console.Wordage.concise)

class BasicTestCase(unittest.TestCase):
    """
    Test Case
    """

    def setUp(self):
        """

        """
        pass


    def tearDown(self):
        """

        """
        pass

    def testWireLog(self):
        """
        Test Class WireLog
        """
        console.terse("{0}\n".format(self.testWireLog.__doc__))

        userDirpath = os.path.join('~', '.ioflo', 'test')
        userDirpath = os.path.abspath(os.path.expanduser(userDirpath))
        if not os.path.exists(userDirpath):
            os.makedirs(userDirpath)

        tempDirpath = tempfile.mkdtemp(prefix="test", suffix="log", dir=userDirpath)
        logDirpath = os.path.join(tempDirpath, 'log')
        if not os.path.exists(logDirpath):
            os.makedirs(logDirpath)

        wire = wiring.WireLog(path=logDirpath)
        prefix = 'localhost'
        midfix = '5000'
        result = wire.reopen(prefix=prefix, midfix=midfix)

        self.assertIsNotNone(wire.rxLog)
        self.assertIsNotNone(wire.txLog)
        self.assertFalse(wire.rxLog.closed)
        self.assertFalse(wire.txLog.closed)
        self.assertTrue(wire.rxLog.name.endswith('_rx.txt'))
        self.assertTrue(wire.txLog.name.endswith('_tx.txt'))
        self.assertTrue("{0}_{1}_".format(prefix, midfix) in wire.rxLog.name)
        self.assertTrue("{0}_{1}_".format(prefix, midfix) in wire.txLog.name)

        ha = ('127.0.0.1', 5000)
        data = b"Test data Tx"
        wire.writeTx(ha, data)
        wire.txLog.seek(0)
        result = wire.txLog.readline()
        result += wire.txLog.readline()
        self.assertEqual(result, ns2b("TX {0}\n{1}\n".format(ha, data.decode('ISO-8859-1'))))

        data = b"Test data Rx"
        wire.writeRx(ha, data)
        wire.rxLog.seek(0)
        result = wire.rxLog.readline()
        result += wire.rxLog.readline()
        self.assertEqual(result, ns2b("RX {0}\n{1}\n".format(ha, data.decode('ISO-8859-1'))))

        wire.close()

        wire = wiring.WireLog(path=logDirpath, same=True)
        prefix = 'localhost'
        midfix = '5000'
        result = wire.reopen(prefix=prefix, midfix=midfix)

        self.assertIsNotNone(wire.rxLog)
        self.assertIsNotNone(wire.txLog)
        self.assertFalse(wire.rxLog.closed)
        self.assertFalse(wire.txLog.closed)
        self.assertIs(wire.rxLog, wire.txLog)
        self.assertEqual(wire.rxLog.name, wire.txLog.name)
        self.assertTrue(wire.rxLog.name.endswith('.txt'))
        self.assertTrue(wire.txLog.name.endswith('.txt'))
        self.assertTrue("{0}_{1}_".format(prefix, midfix) in wire.rxLog.name)
        self.assertTrue("{0}_{1}_".format(prefix, midfix) in wire.txLog.name)

        ha = ('127.0.0.1', 5000)
        data = b"Test data Tx"
        wire.writeTx(ha, data)
        wire.txLog.seek(0)
        result = wire.txLog.readline()
        result += wire.txLog.readline()
        self.assertEqual(result, ns2b("TX {0}\n{1}\n".format(ha, data.decode('ISO-8859-1'))))
        spot = wire.txLog.tell()

        data = b"Test data Rx"
        wire.writeRx(ha, data)
        wire.rxLog.seek(spot)
        result = wire.rxLog.readline()
        result += wire.rxLog.readline()
        self.assertEqual(result, ns2b("RX {0}\n{1}\n".format(ha, data.decode('ISO-8859-1'))))

        wire.close()
        shutil.rmtree(tempDirpath)

    def testWireLogBuffify(self):
        """
        Test Class WireLog
        """
        console.terse("{0}\n".format(self.testWireLogBuffify.__doc__))

        wire = wiring.WireLog(buffify=True)
        prefix = 'localhost'
        midfix = '5000'
        result = wire.reopen(prefix=prefix, midfix=midfix)

        self.assertIsNotNone(wire.rxLog)
        self.assertIsNotNone(wire.txLog)
        self.assertFalse(wire.rxLog.closed)
        self.assertFalse(wire.txLog.closed)

        ha = ('127.0.0.1', 5000)
        data = b"Test data Tx"
        wire.writeTx(ha, data)
        wire.txLog.seek(0)
        result = wire.txLog.readline()
        result += wire.txLog.readline()
        self.assertEqual(result, ns2b("TX {0}\n{1}\n".format(ha, data.decode('ISO-8859-1'))))

        data = b"Test data Rx"
        wire.writeRx(ha, data)
        wire.rxLog.seek(0)
        result = wire.rxLog.readline()
        result += wire.rxLog.readline()
        self.assertEqual(result, ns2b("RX {0}\n{1}\n".format(ha, data.decode('ISO-8859-1'))))

        wire.close()

        wire = wiring.WireLog(buffify=True, same=True)
        prefix = 'localhost'
        midfix = '5000'
        result = wire.reopen(prefix=prefix, midfix=midfix)

        self.assertIsNotNone(wire.rxLog)
        self.assertIsNotNone(wire.txLog)
        self.assertFalse(wire.rxLog.closed)
        self.assertFalse(wire.txLog.closed)
        self.assertIs(wire.rxLog, wire.txLog)

        ha = ('127.0.0.1', 5000)
        data = b"Test data Tx"
        wire.writeTx(ha, data)
        wire.txLog.seek(0)
        result = wire.txLog.readline()
        result += wire.txLog.readline()
        self.assertEqual(result, ns2b("TX {0}\n{1}\n".format(ha, data.decode('ISO-8859-1'))))
        spot = wire.txLog.tell()

        data = b"Test data Rx"
        wire.writeRx(ha, data)
        wire.rxLog.seek(spot)
        result = wire.rxLog.readline()
        result += wire.rxLog.readline()
        self.assertEqual(result, ns2b("RX {0}\n{1}\n".format(ha, data.decode('ISO-8859-1'))))

        wire.close()


def runOne(test):
    '''
    Unittest Runner
    '''
    test = BasicTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)

def runSome():
    """ Unittest runner """
    tests =  []
    names = [
             'testWireLog',
             'testWireLogBuffify',
            ]
    tests.extend(map(BasicTestCase, names))
    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)

def runAll():
    """ Unittest runner """
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(BasicTestCase))
    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:

    #console.reinit(verbosity=console.Wordage.concise)

    #runAll() #run all unittests

    runSome()#only run some

    #runOne('testClientAutoReconnect')





# -*- coding: utf-8 -*-
"""
ioflo test package

To run all the unittests:

from ioflo import test
test.run()

"""
import sys
import unittest
import os


from ioflo.aid.consoling import getConsole
console = getConsole()
console.reinit(verbosity=console.Wordage.concise)

start = os.path.dirname(os.path.dirname
                        (os.path.abspath
                            (sys.modules.get(__name__).__file__)))

# need top to be above root for relative imports to not go above top level
top = os.path.dirname(start)

def run(top, start=None, failfast=False):
    """
    Run unittests starting at directory given by start within the package rooted at top
    """
    if not start:
        start = top

    console.terse("\nRunning ioflo tests starting at '{0}' from '{1}', \n".format(start, top))
    loader = unittest.TestLoader()
    suite = loader.discover(start, 'test_*.py', top )
    unittest.TextTestRunner(verbosity=2, failfast=failfast).run(suite)

if __name__ == "__main__":
    run(top, start)
